/////////////////////////////////////////////////////////////////////////
//	"devtest.cpp"
//------------------------------------------------------------------------
//	$Id: devtest.cpp 126 2004-07-12 10:39:31Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#include	"devtest.h"
#include	"partition.h"
#include	"defs.h"

//////////////////////////////////////////////////////////////////////////
//	DevParam

DevTest::DevParam::DevParam()
{
	this->devname = NULL;
	this->desc = NULL;
	this->inqData = NULL;
	this->capa = NULL;
	this->pathId = 0;
#if		defined(WIN32)
	this->targetId = -1;	// scan
	this->lun      = -1;	// scan
#elif	defined(LINUX)
	this->targetId = 0;
	this->lun      = 0;
#endif
}

DevTest::DevParam::~DevParam()
{
	if ( this->devname != NULL )
		delete [] this->devname;
	if ( this->desc != NULL )
		delete [] this->desc;
}

int		DevTest::DevParam::SetDevName(const char * text)
{
	if ( this->devname != NULL )
	{
		delete [] this->devname;
		this->devname = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->devname = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->devname, text);
	}
	return 0;
}

int		DevTest::DevParam::SetDesc(const char * text)
{
	if ( this->desc != NULL )
	{
		delete [] this->desc;
			this->desc = NULL;
	}
	if ( text != NULL )
	{
		if ( (this->desc = new char [strlen(text)+1]) == NULL )
			return -1;
		strcpy(this->desc, text);
	}
	return 0;
}

char *	DevTest::DevParam::GetDesc(unsigned max)
{
	unsigned	len;

	if ( this->desc == NULL )
		return "";
	if ( max == 0 || (len = strlen(this->desc)) <= max )
		return this->desc;
	return &this->desc[len-max];
}


//////////////////////////////////////////////////////////////////////////
//	DevTest

DevTest::DevTest()
{
	this->opt.cmd            = CMD_SCAN;
#if	defined(WIN32)
	this->opt.scanMode       = SCAN_DRIVE;
#endif
	this->opt.byteUnit       = BYTEUNIT_103;
	this->opt.printPartition = FALSE;
	this->opt.printDetail    = FALSE;
	this->opt.printDevInfo   = FALSE;
	this->opt.dumpModeSense  = FALSE;
	this->opt.benchCommand   = FALSE;
	this->opt.benchSeek      = FALSE;
	this->opt.benchRead      = TRUE;
	this->opt.benchWrite     = FALSE;
	this->opt.benchRandomRW  = FALSE;
	this->opt.benchRwRatio   = 90 * RWRATIO_SCALE / 100;	// 90%
	this->opt.benchCsvResult = NULL;
	this->opt.addr           = I64(0);
	this->opt.count          = I64(0);
	this->opt.readAddr       = I64(0);
	this->opt.writeAddr      = I64(0);
	this->opt.zeroFill       = FALSE;
	this->opt.retry          = 3;
	this->opt.verify         = FALSE;
	this->opt.defBlockSize   = IBM_DISK_BLOCKSIZE;
	this->opt.mode_cdb       = MODE_CDB6;
#if	defined(DEVTEST_DUMP_CSTYLE)
	this->opt.dumpCStyle     = FALSE;
#endif
	this->opt.filename       = NULL;
	this->opt.cpuUsage       = CPUUSAGE_OFF;

	this->ben.mode =
	    BEN_READ
	  | BEN_SEQUENTIAL
	  | BEN_RANDOM
	  | BEN_TRACK_START
	  | BEN_TRACK_END
	  | BEN_BLOCK_MIN
	  | BEN_BLOCK_MAX
	  | BEN_BLOCK_USER;
	this->ben.maxBufferSize   = MAX_BUFFER_SIZE;
	this->ben.testTimeSec     = TEST_TIME_SEC;
	this->ben.randomRWTimeSec = RND_TEST_TIME_SEC;
	this->ben.trackPos        = TRACK_POS_SCALE / 2;
	this->ben.blockSize       = 16384;
	this->ben.seekSector      = 0;	// default
	this->ben.discSpeed       = 0;
#if	defined(WIN32)
	this->ben.useWinFile      = FALSE;
#endif
	this->ben.csvResult       = NULL;

	this->srcDevParam = NULL;
	this->dstDevParam = NULL;
}

DevTest::~DevTest()
{
	if ( this->ben.csvResult != NULL && this->ben.csvResult != stdout && this->ben.csvResult != stderr )
		fclose(this->ben.csvResult);
}

void		help(void);


int		main(int argc, const char ** argv)
{
	int								i;
	DevTest							devTest;
	DevCtl							devctl;
	char							devtmp[256];
	int								cmd = DevTest::CMD_SCAN;
	DevTest::DevArgList				devArgList;
	DevTest::DevArgIterator			it;

	fprintf(stdout, DEVTEST_TITLE   "\n");
	fprintf(stdout, DEVTEST_VERSION "\n");
#if		defined(WIN32)
	devTest.opt.cmd = DevTest::CMD_SCAN;
#elif	defined(LINUX)
	devTest.opt.cmd = DevTest::CMD_SHOW;
#endif

	for ( i = 1; i < argc; i++ )
	{
		const char	*s;

		s = argv[i];
		if ( *s == '-' )
		{
			++s;
			if ( *s == '-' )
			{
				++s;
				if ( strcmpi("210",s) == 0 )
				{												// 210
					devTest.opt.byteUnit = DevTest::BYTEUNIT_210;
				} else if ( strcmpi("103",s) == 0 )
				{												// 103
					devTest.opt.byteUnit = DevTest::BYTEUNIT_103;
				} else if ( strcmpi("benchMark",s) == 0 )
				{												// BENCHMARK
					cmd = devTest.opt.cmd = DevTest::CMD_BENCHMARK;
				} else if ( strcmpi("blockSize",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// blockSize
					s = argv[++i];
					devTest.opt.defBlockSize = atoi(s);
				} else if ( strcmpi("blockTest",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// blockTest
					s = argv[++i];
					devTest.ben.blockSize = atoi(s);
					if ( devTest.ben.blockSize < DevTest::MIN_BUFFER_SIZE || devTest.ben.blockSize >= devTest.ben.maxBufferSize )
					{
						fprintf(stdout, "Blocksize error!\n");
						return -1;
					}
					devTest.ben.mode |= DevTest::BEN_BLOCK_USER;
				} else if ( strcmpi("cdspeed",s) == 0 )
				{												// cdspeed
					devTest.ben.discSpeed = DevTest::CD_SPEED;
				} else if ( strcmpi("count",s) == 0 && i + 1 < argc && (isdigit(argv[i+1][0]) || strcmpi(argv[i+1],"max") == 0 ) )
				{												// count
					s = argv[++i];
					if ( strcmpi(s, "max") == 0 )
					{
						devTest.opt.count = 0;
					} else 
					{
						devTest.opt.count = _atoi64(s);
						if ( devTest.opt.count <= 0 )
						{
							fprintf(stdout, "Too small count!\n");
							return -1;
						}
					}
				} else if ( strcmpi("cpuUsage",s) == 0 )	// cpuUsage
				{
					devTest.opt.cpuUsage = DevTest::CPUUSAGE_SELF;
					if ( i + 1 < argc && argv[i+1][0] != '-' )
					{
						if ( strcmpi(argv[i+1], "self") == 0 )
						{
							++i;
							devTest.opt.cpuUsage = DevTest::CPUUSAGE_SELF;
						}
#if	defined(WIN32)
						else if ( strcmpi(argv[i+1], "all") == 0 )
						{
							++i;
							devTest.opt.cpuUsage = DevTest::CPUUSAGE_ALL;
						}
#endif
#if	defined(LINUX)
						else if ( strcmpi(argv[i+1], "sys") == 0 )
						{
							++i;
							devTest.opt.cpuUsage = DevTest::CPUUSAGE_SYS;
						} else if ( strcmpi(argv[i+1], "user") == 0 )
						{
							++i;
							devTest.opt.cpuUsage = DevTest::CPUUSAGE_USER;
						}
#endif
					}
				} else if ( strcmpi("csvresult",s) == 0 )
				{												// csvresult
					if ( i + 1 < argc && argv[i+1][0] != '-' )
						devTest.opt.benchCsvResult = argv[++i];
					else
						devTest.opt.benchCsvResult = "stdout";
				} else if ( strcmpi("dump",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// DUMP
					s = argv[++i];
					devTest.opt.addr = _atoi64(s);
					cmd = devTest.opt.cmd = DevTest::CMD_DUMP;
#if	defined(DEVTEST_DUMP_CSTYLE)
				} else if ( strcmpi("dumpCStyle",s) == 0 )
				{												// dumpCStyle
					devTest.opt.dumpCStyle = TRUE;
#endif
				} else if ( strcmpi("dvdspeed",s) == 0 )
				{												// dvdspeed
					devTest.ben.discSpeed = DevTest::DVD_SPEED;
				} else if ( strcmpi("file",s) == 0 && i + 1 < argc && argv[i+1][0] != '-' )
				{												// file
					s = argv[++i];
					devTest.opt.filename = s;
				} else if ( strcmpi("maxBufferSize",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// maxBufferSize
					s = argv[++i];
					devTest.ben.maxBufferSize = atoi(s);
					if ( devTest.ben.maxBufferSize < DevTest::MIN_BUFFER_SIZE )
					{
						fprintf(stdout, "Too small MaxBufferSize!\n");
						return -1;
					}
				} else if ( strcmpi("mode6",s) == 0 )			// mode6
				{
					devTest.opt.mode_cdb = DevTest::MODE_CDB6;
				} else if ( strcmpi("mode10",s) == 0 )			// mode10
				{
					devTest.opt.mode_cdb = DevTest::MODE_CDB10;
				} else if ( strcmpi("modeSenseDump",s) == 0 )
				{												// modeSenseDump
					devTest.opt.dumpModeSense = TRUE;
				} else if ( strcmpi("readblock",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// ReadBlock
					s = argv[++i];
					devTest.opt.addr = _atoi64(s);
					cmd = devTest.opt.cmd = DevTest::CMD_READBLOCK;
					if ( devTest.opt.count == 0 )
						devTest.opt.count = 1;
				} else if ( strcmpi("retry",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// retry
					s = argv[++i];
					devTest.opt.retry = atoi(s);
				} else if ( strcmpi("randomRWTime",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// randomRWTime
					s = argv[++i];
					devTest.ben.randomRWTimeSec = atoi(s);
					if ( devTest.ben.randomRWTimeSec < 1 )
					{
						fprintf(stdout, "Too small rndTestTime!\n");
						return -1;
					}
				} else if ( strcmpi("seekSector",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// seekSector
					s = argv[++i];
					devTest.ben.seekSector = atoi(s);
				} else if ( strcmpi("surfscan",s) == 0 )
				{												// SURFSCAN
					cmd = devTest.opt.cmd = DevTest::CMD_SURFSCAN;
					if ( i + 1 < argc && isdigit(argv[i+1][0]) )
					{
						++i;
						devTest.opt.addr = _atoi64(argv[i]);
					}
				} else if ( strcmpi("testTime",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// testTime
					s = argv[++i];
					devTest.ben.testTimeSec = atoi(s);
					if ( devTest.ben.testTimeSec < 1 )
					{
						fprintf(stdout, "Too small testTime!\n");
						return -1;
					}
				} else if ( strcmpi("trackZone",s) == 0 && i + 1 < argc && isdigit(argv[i+1][0]) )
				{												// trackZone
					s = argv[++i];
					devTest.ben.trackPos = DevTest::TRACK_POS_SCALE * atoi(s) / 100;
					if ( devTest.ben.trackPos < 0 || devTest.ben.trackPos >= DevTest::TRACK_POS_SCALE )
					{
						fprintf(stdout, "Track Zone error!\n");
						return -1;
					}
					devTest.ben.mode |= DevTest::BEN_TRACK_USER;
#if	defined(WIN32)
				} else if ( strcmpi("winfile",s) == 0 )
				{												// winfile
					devTest.ben.useWinFile = TRUE;
#endif
				} else if ( strcmpi("verify",s) == 0 )
				{												// verify
					devTest.opt.verify = TRUE;
				} else
				{
					fprintf(stderr, "Parameter error! (%d:%s)\n", i, argv[i]);
					return 1;
				}
			} else
			{
				while ( *s )
				{
					switch ( toupper(*s) )
					{
						case '?':
						case 'H':
							help();
							return 1;
						case 'P':
							devTest.opt.printPartition = TRUE;
							break;
						case 'D':
							devTest.opt.printDevInfo = TRUE;
							break;	// Add (Bug Fix)
						case 'V':
							devTest.opt.printDetail = TRUE;
							break;
						case 'C':
							devTest.opt.benchCommand = TRUE;
							break;
						case 'S':
							devTest.opt.benchSeek = TRUE;
							break;
						default:
							fprintf(stderr, "Parameter error! (%d:%s '%c')\n", i, argv[i], *s);
							return 1;
					}
					++s;
				}
			}
		} else
		{
			DevTest::DevParam	*devParam;

			if ( (devParam = new DevTest::DevParam) == NULL )
				return 1;
			devParam->SetDesc(s);
			if ( i + 1 < argc && isdigit(argv[i+1][0]) )
			{	// <device> <scsi-id>[:<lun>]
				s = argv[++i];
				devParam->targetId = atoi(s);
				devParam->lun = 0;
				while ( *s )
				{	// host adapter #
					if ( isdigit(*s) == 0 )
						break;
					++s;
				}
				if ( *s == ':' || *s == '-' )
				{	// lun
					++s;
					if ( isdigit(*s) )
						devParam->lun = atoi(s);
				}
			}
			devArgList.push_back(devParam);
		}
	}
#if	defined(WIN32)
	if ( devArgList.size() == 0 )
	{
		DevTest::DevParam	*devParam;

		if ( (devParam = new DevTest::DevParam) == NULL )
			return 1;
		devArgList.push_back(devParam);
	}
#endif

	for ( it = devArgList.begin(); it != devArgList.end(); it++ )
	{
		DevTest::DevParam		*devParam;
		BOOL					setDevName;

		devParam = *it;
		setDevName = FALSE;
		if ( devTest.srcDevParam == NULL )
			devTest.srcDevParam = devParam;
		else if ( devTest.dstDevParam == NULL )
			devTest.dstDevParam = devParam;


#if	defined(WIN32)
		if ( devParam->desc == NULL )
		{	// Scan

		} else if ( (devParam->desc[1] == ':' || devParam->desc[1] == '\0') ) // drive letter
		{
			if ( isalpha(devParam->desc[0]) )
			{
				sprintf(devtmp, "\\\\.\\%c:", toupper(devParam->desc[0]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			//	devTest.opt.cmd = DevTest::CMD_SHOW;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_DRIVE;
			}
		} else if ( _strnicmp(devParam->desc, "disk", 4) == 0 )	// disk
		{
			if ( isdigit(devParam->desc[4]) )
			{
				sprintf(devtmp, "\\\\.\\PhysicalDrive%d", (int)atoi(&devParam->desc[4]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_DISK;
			}
		} else if ( _strnicmp(devParam->desc, "scsi", 4) == 0 )	// scsi
		{
			if ( isdigit(devParam->desc[4]) )
			{
				sprintf(devtmp, "\\\\.\\scsi%d:", (int)atoi(&devParam->desc[4]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
		//		devTest.opt.cmd = DevTest::CMD_SHOW;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_SCSI;
			}
#if	defined(DEVTEST_USE_CDROM)
		} else if ( _strnicmp(devParam->desc, "cdrom", 5) == 0 )	// CD-ROM
		{
			if ( isdigit(devParam->desc[5]) )
			{
				sprintf(devtmp, "\\\\.\\CdRom%d", (int)atoi(&devParam->desc[5]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_CDROM;
			}
#endif
		} else
		{
			sprintf(devtmp, "\\\\.\\%s", devParam->desc);
			devParam->SetDevName(devtmp);
			setDevName = TRUE;
		}
		if ( setDevName != FALSE )
		{
			if ( devTest.opt.cmd == DevTest::CMD_SCAN )
				devTest.opt.cmd = DevTest::CMD_SHOW;
		} else
			devTest.opt.cmd = DevTest::CMD_SCAN;
#endif // end if ifdef WIN32

#if	defined(LINUX)
		if ( devParam->desc == NULL )
			continue;
		devParam->SetDevName(devParam->desc);
#endif

		switch ( devTest.opt.cmd )
		{
			case DevTest::CMD_SCAN:
#if	defined(WIN32)
				devTest.ScanDevice(*devParam);
#endif
				break;

			case DevTest::CMD_SHOW      :
			case DevTest::CMD_DUMP      :
			case DevTest::CMD_BENCHMARK :
			case DevTest::CMD_SURFSCAN  :
			case DevTest::CMD_READBLOCK :
				if ( devParam->devname == NULL )
					return 1;
				devTest.TargetDevice(*devParam);
				break;
		}
	} // end of for

	return 0;
}

void	help(void)
{
	static const char *msg[] =
	{
		"usage: devtest [<device>] [<option>]",
#if	defined(WIN32)
		"  <device>",
		"    <c>:               ... Drive Letter      (ex. c:)    : Scan ?:",
		"    disk[<n>]          ... Physical Drive    (ex. disk0) : without <n> Scan",
#if	defined(DEVTEST_USE_CDROM)
		"    cdrom[<n>]         ... CD/DVD Dtive      (ex. cdrom0): without <n> Scan",
#endif
		"    scsi[<n>]          ... SCSI Host Adapter (ex. scsi0) : without <n> Scan",
#endif
		"  <command>",
		"    --benchmark        ... Benchmark test",
		"    --dump <adr>       ... Dump Sector",
		"    --readBlock  <adr> ... Read Block/Sector Data",
		"    --surfscan [<adr>] ... Surface Scan",
		"  <option>",
		"    -c                 ... Command test      (Benchmark",
		"    -d                 ... Print Device Information (Scan, Show",
		"    -p                 ... Print Partition   (Scan, Show",
		"    -s                 ... Seek test         (BenchMark",
		"    -v                 ... Print Detail      (Scan, Show",
		"    --103              ... k = 10^3 Bytes (default)",
		"    --210              ... K = 2^10 Bytes",
		"    --blockSize <n>    ... Default Block(Sector) Size",
		"    --blockTest <n>    ... Set Blocksize     (Benchmark",
		"    --cdspeed          ... Report CD speed   (Benchmark",
		"    --cpuUsage [<mode>]... Report CPU %      (Benchmark,SurfScan",
#if	defined(WIN32)
		"                           mode: self, all",
#elif	defined(LINUX)
		"                           mode: self, sys, user",
#endif
		"    --count <n>        ... Block R/W count, n>=1 or max (R/W Block",
		"    --dvdspeed         ... Report DVD speed  (Benchmark",
		"    --file <filename>  ... Read/Write file   (R/W Block",
		"    --maxBufferSize <n>... Max Buffer Size   (Benchmark,Surfscan",
		"    --seekSector <n>   ... Seek Sector/Track (Benchmark Seek Test",
		"    --testTime <sec>   ... Test Time         (Benchmark",
		"    --trackZone <per>  ... Track Zone 1-99%  (Benchmark",
		"    --retry <n>        ... Error Retry Count (Surfscan",
		"    --verify           ... Verify Comamnd    (Surfscan",
#if	defined(WIN32)
		"    --winfile          ... Use Windows API   (Benchmark",
#endif
#if	defined(WIN32)
		"  Scan     : devtest [<device>] [<option>]",
#endif
		"  Show     : devtest  <device>  [<option>]",
		"  Benchmark: devtest  <device> --benchmark [<option>]",
		"  Dump     : devtest  <device> --dump <adr> [<option>]",
		"  Surfscan : devtest  <device> --surfscan [<adr>] [<option>]",
	};
	int			i;

	for ( i = 0; i < sizeof(msg)/sizeof(*msg); i++)
		puts(msg[i]);
}

#if	defined(WIN32)
//////////////////////////////////////////////////////////////////////////
//	Scan Device

int		DevTest::ScanDevice(DevTest::DevParam & devParam)
{
	switch ( this->opt.scanMode )
	{
		case SCAN_DRIVE:
			this->ScanDeviceDrive(devParam);
			break;
		case SCAN_DISK:
			this->ScanDeviceDisk(devParam);
			break;
		case SCAN_SCSI:
			this->ScanDeviceScsi(devParam);
			break;
#if	defined(DEVTEST_USE_CDROM)
		case SCAN_CDROM:
			this->ScanDeviceCdrom(devParam);
			break;
#endif
	}
	return 0;
}

// drive letter
int		DevTest::ScanDeviceDrive(DevTest::DevParam & devParam)
{
	int		drv;
	char	devname[MAX_PATH], desc[16];

	for ( drv = 'A'; drv <= 'Z'; drv++ )
	{
		sprintf(desc, "%c:", drv);
		sprintf(devname, "\\\\.\\%c:", drv);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		this->ScanDeviceCheck(devParam);
	}

	return 0;
}

// disk (physical drive
int		DevTest::ScanDeviceDisk(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "disk%d", n);
		sprintf(devname, "\\\\.\\PhysicalDrive%d", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}

// SCSI
int		DevTest::ScanDeviceScsi(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "scsi%d", n);
		sprintf(devname, "\\\\.\\scsi%d:", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}


#if	defined(DEVTEST_USE_CDROM)
// CD-ROM
int		DevTest::ScanDeviceCdrom(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "cdrom%d", n);
		sprintf(devname, "\\\\.\\CdRom%d", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}
#endif

int		DevTest::ScanDeviceCheck(DevTest::DevParam & devParam)
{
	DevCtl::ScsiDevData 	*scsiDev = NULL;
	SCSI_ADDRESS			scsiAddr;
	BOOL					target = FALSE;

	if ( devParam.devname == NULL )
		return -1;
	this->devCtl.CloseDevice();
	devParam.inqData = NULL;
	devParam.capa    = NULL;
	if ( this->devCtl.OpenDevice(devParam.devname, DevCtl::RW_READ | DevCtl::RW_WRITE | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE) < 0 )
		return -1;	// error

	if ( this->devCtl.GetHandleType() != DevCtl::HT_SCSI )
	{
		if ( this->devCtl.ScsiGetAddress(scsiAddr) == 0 )
			target = TRUE;
	}
	if ( (scsiDev = this->devCtl.ScsiGetInquiryData()) != NULL )
	{
		int			i;

		for ( i = 0; i < (int)scsiDev->GetBusNum(); i++ )
		{
			DevCtl::ScsiBusData 	*busData;
			int						j;

			if ( (busData = scsiDev->GetBusData(i)) == NULL )
				continue;
			for ( j = 0; j < (int)busData->GetInqNum(); j++ )
			{
				DevCtl::ScsiInqData 	*inqData;

				if ( (inqData = busData->GetInqData(j)) == NULL )
					continue;
				if ( target == FALSE || (target != FALSE && scsiAddr.TargetId == inqData->GetTargetId() && scsiAddr.Lun == inqData->GetLun()) )
				{
					if ( this->opt.printPartition || this->opt.printDevInfo )
						fprintf(stdout,"\n");
					if ( target == FALSE && (i != 0 || j != 0) )
						devParam.SetDesc(NULL);
					devParam.inqData = inqData;
					this->PrintDevice(devParam);
				}
			}
		}
		delete scsiDev;
	} else
	{
		DevCtl::ScsiInqData	inqData;
		unsigned			pathId, targetId, lun;

		if ( this->devCtl.GetFileType() != FILE_TYPE_DISK )
			return -1;
		if ( target == TRUE )
		{
			pathId   = scsiAddr.PathId;
			targetId = scsiAddr.TargetId;
			lun      = scsiAddr.Lun;
		} else
		{
			pathId = targetId = lun = 0;
		}
		if ( this->devCtl.ScsiCreateInqData(inqData, pathId, targetId, lun) < 0 )
			return -1;
		inqData.unknownId = TRUE;
		devParam.inqData = &inqData;
		this->PrintDevice(devParam);
	}

	return 0;
}
#endif

int	DevTest::OpenTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam)
{
	int						pathId, targetId, lun;

#if	defined(WIN32)
	tg.scsiDev = NULL;
	tg.busData = NULL;
#endif
	tg.inqData    = NULL;
	tg.setInqData = FALSE;

	if ( devParam.devname == NULL )
		return -1;
	pathId   = devParam.pathId;
	targetId = devParam.targetId;
	lun      = devParam.lun;
	devCtl.CloseDevice();
#if	defined(LINUX)
	devCtl.SetDefaultBlockSize(this->opt.defBlockSize);
#endif
	devParam.inqData = NULL;
	devParam.capa    = NULL;
	if ( devCtl.OpenDevice(devParam.devname
	  , DevCtl::RW_READ | DevCtl::RW_WRITE | DevCtl::RW_SYNC | DevCtl::RW_NONBLOCK | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE) < 0 )
	{
		if ( devCtl.OpenDevice(devParam.devname
		  , DevCtl::RW_READ | DevCtl::RW_SYNC | DevCtl::RW_NONBLOCK | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE) < 0 )
		{
			fprintf(stderr,"%s - Device open error!\n", devParam.devname);
			return -1;	// error
		}
	}

#if	defined(WIN32)
	if ( pathId < 0 || targetId < 0 || lun < 0 )
	{
		SCSI_ADDRESS			scsiAddr;

		if ( devCtl.ScsiGetAddress(scsiAddr) == 0 )
		{	// Get SCSI Address
			pathId   = scsiAddr.PathId;
			targetId = scsiAddr.TargetId;
			lun      = scsiAddr.Lun;
		} else
		{
			pathId = targetId = lun = 0;
			if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) < 0 )
			{	// error or Drive not ready
				fprintf(stderr,"%s - Device error!\n", devParam.GetDesc());
				return -1;
			}
			tg.inqData = &tg.inqDataWork;
			tg.inqData->unknownId = TRUE;
			tg.setInqData = TRUE;
		}
	}
	if ( tg.setInqData == FALSE )
	{
		if ( (tg.scsiDev = devCtl.ScsiGetInquiryData()) != NULL )
		{
			if ( (tg.busData = tg.scsiDev->GetBusData(pathId)) == NULL )
				return -1;
			for ( int j = 0; j < (int)tg.busData->GetInqNum(); j++ )
			{
				if ( (tg.inqData = tg.busData->GetInqData(j)) == NULL )
					continue;
				if ( (int)tg.inqData->GetPathId()   == pathId
				  && (int)tg.inqData->GetTargetId() == targetId
				  && (int)tg.inqData->GetLun()      == lun )
				{
					tg.setInqData = TRUE;
					break;
				}
			}
		} else
		{
			if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) < 0 )
				return -1;
			tg.inqData = &tg.inqDataWork;
			tg.inqData->unknownId = TRUE;
			tg.setInqData = TRUE;
		}
	}
#endif
#if	defined(LINUX)
	{
		if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) < 0 )
			return -1;
		tg.inqData = &tg.inqDataWork;
		tg.inqData->unknownId = FALSE;
		tg.setInqData = TRUE;
	}
#endif

	if ( tg.setInqData != FALSE && tg.inqData != NULL )
	{
		devParam.inqData = tg.inqData;
	} else
		return -1;

	return 0;
}

void	DevTest::CloseTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam )
{
#if	defined(WIN32)
	if ( tg.scsiDev != NULL )
	{
		delete tg.scsiDev;
		tg.scsiDev = NULL;
	}
	devParam.inqData = NULL;
#endif
}

//////////////////////////////////////////////////////////////////////////
//	Show,Dump,Benchmark,Surfscan: In DevTest.opt.cmd

int		DevTest::TargetDevice(DevTest::DevParam & devParam)
{
	DevTargetParam			tg;

	switch ( this->opt.cmd )
	{
		case CMD_SHOW     :
		case CMD_DUMP     :
			break;
		case CMD_BENCHMARK:
			if ( this->opt.benchCsvResult != NULL && this->ben.csvResult == NULL )
			{
				if ( strcmp(this->opt.benchCsvResult, "stdout") == 0 )
				{
					this->ben.csvResult = stdout;
				} else if ( strcmp(this->opt.benchCsvResult, "stderr") == 0 )
				{
					this->ben.csvResult = stderr;
				} else
				{
					if ( (this->ben.csvResult = fopen(this->opt.benchCsvResult, "a")) == NULL )
					{
						fprintf(stderr, "File open error! (%s)\n", this->opt.benchCsvResult);
						return -1;
					}
				}
			}
			break;
		case CMD_SURFSCAN :
			break;
		case CMD_READBLOCK :
			break;
		default:
			return -1;
	}
	if ( this->OpenTargetDevice(this->devCtl, tg, devParam) == 0 )
	{
		switch ( this->opt.cmd )
		{
			case CMD_SHOW    :
				this->PrintDevice(devParam);
				break;
			case CMD_DUMP    :
				this->PrintDump(devParam);
				break;
			case CMD_BENCHMARK:
			{
				fprintf(stdout, "***** BENCHMARK TEST *****\n\n");

				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					this->InitCpuUsage();
				this->opt.printPartition = FALSE;
				this->opt.printDetail    = FALSE;
				this->PrintDevice(devParam);
				this->BenchCommand( devParam, 0);
				this->BenchSeek( devParam, 0);
				this->BenchReadWrite( devParam, (this->ben.mode & ~BEN_WRITE) | BEN_READ);
				// Delete Source Code: Write Benchmark
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					this->TermCpuUsage();
				break;
			}
			case CMD_SURFSCAN:
			{
				this->SurfaceScan(devParam);
				break;
			}
			case CMD_READBLOCK:
				this->ReadWriteBlock(devParam);
				break;

		}
	}
	this->CloseTargetDevice(this->devCtl, tg, devParam);

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	ShowDevice

int		DevTest::PrintDevice(DevTest::DevParam & devParam)
{
	BYTE						*data;
	DevCtl::SCSIINQUIRY 		*pinq;
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;
	char						sizeBuf[80];

	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	data     = devParam.inqData->GetData();
	pinq     = devParam.inqData->GetInquiryData();	// data == pinq
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{	// error
		devParam.capa = NULL;
		blkNum = 0;
		blkLen = 0;
		sizeBuf[0] = '\0';
	} else
	{
		devParam.capa = &capa;
		blkNum =           this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
		blkLen = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
		this->FormatSize(sizeBuf,  blkNum * blkLen);
	}

	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "###### #-#-# ############################ - ####################### ### #######"
	if ( strlen(devParam.GetDesc()) > 6 )
		fprintf(stdout, "Device: %s\n", devParam.GetDesc());
	{
		char	msg[256];

		sprintf(msg, "%-6s %X-%X-%X %.28s - %-23s %3s %7s\n"
		  , devParam.GetDesc(6)
		  , pathId
		  , targetId
		  , lun
		  , pinq->venderId
		  , (data[2] & 3) == 0 ? "" :  this->devCtl.ScsiGetDeviceType(data[1])
		  , (data[1] & 0x80) ? "RMB" : "FIX"
		  , sizeBuf );
		fprintf(stdout, "%s", msg);
		if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
			fprintf(this->ben.csvResult, "%s", msg);
	}
	if ( devParam.capa != NULL
	  && this->opt.printPartition
#if		defined(WIN32)
	  && (  this->devCtl.GetHandleType() == DevCtl::HT_DISK
	     || this->devCtl.GetHandleType() == DevCtl::HT_DRIVE )
#elif	defined(LINUX)
	  && (   this->devCtl.GetHType() == DevCtl::HT_BLKDEV
		  || this->devCtl.GetHType() == DevCtl::HT_CDROM
	      || (pinq->devTypeModify&0x1F) == 0x00	// Direct Access Device
	      || (pinq->devTypeModify&0x1F) == 0x05	// CD/DVD Device
	     )
#endif
	)
	{	// Partition
		this->PrintPartition(devParam);
	}
	if ( this->opt.printDevInfo )
	{	// Device
		this->PrintDevInfo(devParam);
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Print Partition Table

int		DevTest::PrintPartition(DevTest::DevParam & devParam)
{
	BYTE						*buffer = NULL;
	INT64						blkNum;
	unsigned					blkLen;
	DevCtl::SCSICAPACITYDATA	capaBuf;
	unsigned					pathId, targetId, lun, blockSize;
	BOOL						cdrom = FALSE;

	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( devParam.capa == NULL || devParam.capa->blockLength == 0 )
	{
		if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capaBuf), &capaBuf) < 0 )
			return -1;
		devParam.capa = &capaBuf;
	}
	// CD-ROM
	if ( this->PrintCdrom(devParam) == 0 )
		cdrom = TRUE;

	blkNum =           this->devCtl.ScsiGetLBA(devParam.capa->logicalBlockAddress);
	blkLen = (unsigned)this->devCtl.ScsiGetLBA(devParam.capa->blockLength);
	if ( blkLen > DevTest::SECTOR_SIZE_MAX )
		return -1;
	blockSize = blkLen;
	if ( (buffer = new BYTE [blockSize]) == NULL )
		return -1;
	if ( this->devCtl.ScsiRead(pathId, targetId, lun, I64(0), 1, blockSize, buffer) < (int)blockSize )
		return -1;

	if ( blockSize >= IBM_DISK_BLOCKSIZE && buffer[IBM_DISK_BLOCKSIZE-2] == 0x55 && buffer[IBM_DISK_BLOCKSIZE-1] == 0xAA )
	{	// PC/AT Windows
		if ( buffer[0] == 0xE9 || buffer[0] == 0xEB )
		{	// IBM floppy
			BPBDATA			*bpb;
			char			serial[80], fileSys[80];
			INT64			secVol;

			// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
			DevTest::fprintf(stdout, "  IBM FLOPPY-DISK FORMAT  (%8u) %10"FI64"u sectors   %20,I64uB\n"
			  , blockSize
			  , (INT64)blkNum
			  , (INT64)blkNum * (INT64)blockSize
			);
			bpb = (BPBDATA *) &buffer[0x0B];
			if ( bpb->wRootEntry > 0 || bpb->wFatSector )
			{
				sprintf(serial, "%04X-%04X", bpb->dwSerialNo >> 16, bpb->dwSerialNo & 0xFFFF);
				sprintf(fileSys, "%.8s", bpb->sFileSystem);
				if ( (secVol = (INT64)bpb->wSectorVolume) == 0 )
					secVol = (INT64)bpb->dwLongSectorVolume;
				fprintf(stdout,"       %.11s %.9s %13"FI64"u .. %-24s %.8s\n"
				  , bpb->sVolumeLabel
				  , serial
				  , (INT64)secVol
				  , fileSys
				  , &buffer[0x03]
				);
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, I64(0));
			} else
			{	// FAT32
				BF_BPB	*bfbpb;

				bfbpb = (BF_BPB *)bpb;
				sprintf(serial, "%04X-%04X", bfbpb->dwSerialNo >> 16, bfbpb->dwSerialNo & 0xFFFF);
				sprintf(fileSys, "%.8s", bfbpb->sFileSystem);
				if ( (secVol = (INT64)bfbpb->wSectorVolume) == 0 )
					secVol = (INT64)bpb->dwLongSectorVolume;
				fprintf(stdout,"       %.11s %.9s %13"FI64"u .. %-24s %.8s\n"
				  , bfbpb->sVolumeLabel
				  , serial
				  , (INT64)secVol
				  , fileSys
				  , &buffer[0x03]
				);
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, I64(0));
			}
		} else if ( blockSize == IBM_DISK_BLOCKSIZE )
		{	// IBM HARDDISK
			int				ext;
			INT64			extAddr;

			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "###### #-#-# ############################ - ####################### ### #######"
			// "### [### ## ##-### ## ##] ########## ########## ##:#################### #######
			// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
			DevTest::fprintf(stdout, "  IBM HARD-DISK FORMAT    (%8u) %10"FI64"u sectors   %20,I64uB\n"
			  , blockSize
			  , (INT64)blkNum
			  , (INT64)blkNum * (INT64)blockSize
			);
			extAddr = 0;
			ext = 0;
			fprintf(stdout, "No  [Start CHS- End  CHS]   StartSec  SectorNum HH:Partition Type          SIZE\n");
			fprintf(stdout, "--- --------------------- ---------- ---------- --:-------------------- -------\n");
			this->PrintIbmPartition(devParam, ext, extAddr, I64(0), buffer);
		}
	} else if ( buffer[0] == 'I' && buffer[1] == 'P' && buffer[2] == 'L' )
	{	// FMR
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "  FMR HARD-DISK FORMAT    (########) ########## sectors   ####################B
		DevTest::fprintf(stdout, "  FMR HARD-DISK FORMAT    (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
	} else 
	{
#if	defined(DEVTEST_USE_UDFLIB)
		if ( this->PrintDisc(devParam) == 0 )
		{
		//	fprintf(stdout, "  UDF FORMAT DISC\n");
			cdrom = TRUE;
		}
#endif
		if ( cdrom == FALSE )
		{
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "  UNKNOWN FORMAT          (########) ########## sectors   ####################B
		DevTest::fprintf(stdout, "  UNKNOWN FORMAT          (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
		}
	}
	return 0;
}

int		DevTest::PrintIbmPartition(DevTest::DevParam & devParam, int & ext, INT64 & extAddr, INT64 baseAddr, BYTE * buffer)
{
	int				partNo;
	IBMPARTINF		partinf[4];
	INT64			blkNum;
	unsigned		blkLen;
	unsigned		pathId, targetId, lun, blockSize;

	if ( devParam.capa == NULL )
		return -1;
	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	blkNum =           this->devCtl.ScsiGetLBA(devParam.capa->logicalBlockAddress);
	blkLen = (unsigned)this->devCtl.ScsiGetLBA(devParam.capa->blockLength);
	blockSize = blkLen;

	memcpy( &partinf, buffer + 0x1BE, sizeof(partinf));
	for ( partNo = 0; partNo < 4; partNo++ )
	{
		BOOL			err;
		IBMPARTINF	   *ppart;
		char			sizeBuf[32], noBuf[8], partName[80];

		err = 0;
		ppart = &partinf[partNo];
		if ( ppart->dwRelativeSector == 0 )
			continue;
		this->FormatSize(sizeBuf,  (INT64)ppart->dwSectorNumber * blkLen);
		if ( ext == 0 )
			sprintf(noBuf, "%1d ", partNo + 1);
		else if ( ext < 9 )
			sprintf(noBuf, "e%1d", ext );
		else
			sprintf(noBuf, "%2d", ext );
		if ( (ppart->bPartType == 0x05 || ppart->bPartType == 0x0F) && ext > 0 )
			sprintf(partName, "Extended  %10"FI64"u", (INT64)baseAddr);
		else
			strcpy(partName, this->GetPartitionType(ppart->bPartType));
		if ( extAddr + ppart->dwRelativeSector >= blkNum
		  || extAddr + ppart->dwRelativeSector + ppart->dwSectorNumber > blkNum )
		{
			sprintf(partName, "Error     %10"FI64"u", (INT64)baseAddr);
			err = 1;
		}
		if ( (ppart->bPartType == 0x05 || ppart->bPartType == 0x0F) && ext && this->opt.printDetail == FALSE )
		{	// extended partition
			
		} else
		{
			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "### [### ## ##-### ## ##] ########## ########## ##:#################### #######
			fprintf(stdout, "%1s%2s [%03X %02X %02X-%03X %02X %02X] %10u %10u %02X:%-20s %7s\n"
			  , ppart->bBootFlag ? "*" : "" 
			  , noBuf
			  , ppart->bStart[2] + ((ppart->bStart[1]&0xC0)<<2)
			  , ppart->bStart[0]
			  , ppart->bStart[1]&0x3F
			  , ppart->bEnd[2] + ((ppart->bEnd[1]&0xC0)<<2)
			  , ppart->bEnd[0]
			  , ppart->bEnd[1]&0x3F
			  , ppart->dwRelativeSector + (unsigned)baseAddr
			  , ppart->dwSectorNumber
			  , ppart->bPartType
			  , partName // err == 0 && ext == 0 ? this->GetPartitionType(ppart->bPartType) : errPart
			  , sizeBuf
			);
		}
		switch ( ppart->bPartType )
		{
			case 0x05:	/* extend partition	*/
			case 0x0F:	/* extend partition	*/
			{
				INT64	addr;

				if ( extAddr + ppart->dwRelativeSector >= blkNum )
					break;	// error
				addr = ppart->dwRelativeSector + extAddr;
				if ( extAddr == 0 )
					extAddr = addr;
				if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr, 1, blockSize, buffer) < (int)blockSize )
					break;	// error
				ext++;
				this->PrintIbmPartition(devParam, ext, extAddr, addr, buffer);
				ext--;
				break;
			}
			case 0x01:	// FAT12
			case 0x04:	// FAT16(<32M)
			case 0x06:	// FAT16(>=32M)
			case 0x0B:	// FAT32
			case 0x0C:	// FAT32 LBA
			case 0x0E:	// FAT16 LBA
			case 0x11:	// Hidden FAT12
			case 0x14:	// Hidden FAT16 (<32M)
			case 0x16:	// Hidden FAT16 (>=32M)
			case 0x1B:	// Hidden FAT32 Hidden
			case 0x1C:	// Hidden FAT32 LBA
			case 0x1E:	// Hidden FAT16 LBA
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, ppart->dwRelativeSector + baseAddr);
				break;

			case 0x07:	// NTFS
				if ( this->opt.printDetail )
					this->PrintNtfsInfo(devParam, ppart->dwRelativeSector + baseAddr);
				break;

			case 0x83:	// Linux ext2
				if ( this->opt.printDetail )
					this->PrintExt2Info(devParam, ppart->dwRelativeSector + baseAddr, ppart->dwSectorNumber);
				break;
		}
	}
	return 0;
}

int		DevTest::PrintFatInfo(DevTest::DevParam & devParam, INT64 addr)
{
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;
	BYTE						*buffer = NULL;
	int							retcode;

	if ( devParam.capa == NULL )
		return -1;
	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	blkNum   =           this->devCtl.ScsiGetLBA(devParam.capa->logicalBlockAddress);
	blkLen   = (unsigned)this->devCtl.ScsiGetLBA(devParam.capa->blockLength);
	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	try
	{
		char			t1[256], t2[256];
		void *			p;
	
		if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr, 1, blkLen, buffer) < (int)blkLen )
			throw -1;
		p = &buffer[0x0B];
		if ( ((BPBDATA *)p)->wRootEntry > 0 || ((BPBDATA *)p)->wFatSector )
		{	// FAT16
			BPBDATA			*bpb;

			bpb = (BPBDATA *)p;
			PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", buffer[0],buffer[1],buffer[2]), 0);
			PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", &buffer[3]), 0);
			PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorSize), 0);
			PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", bpb->bAllocUnit, (unsigned)bpb->bAllocUnit * bpb->wSectorSize), 0);
			PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->wReserveSector), 0);
			PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", bpb->bFatNumber), 0);
			PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", bpb->wRootEntry), 0);
			PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->wSectorVolume, (INT64)bpb->wSectorSize * bpb->wSectorVolume), 0);
			PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", bpb->bMediaDescriptor), 0);
			PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", bpb->wFatSector), 0);
			PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorPerTrack), 0);
			PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", bpb->wHeadNumber), 0);
			PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", bpb->dwHiddenSector), 0);
			PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->dwLongSectorVolume, (INT64)bpb->wSectorSize * bpb->dwLongSectorVolume), 0);
			PutPos( 4, "Serial No"         , COLUMN1+1, Fmt(t1,"%04X-%04X", bpb->dwSerialNo>>16, bpb->dwSerialNo & 0xFFFF), 0);
			PutPos( 4, "Volume Label"      , COLUMN1-1, Fmt(t1,"%11.11s", bpb->sVolumeLabel), 0);
			PutPos( 4, "File System"       , COLUMN1+2, Fmt(t1,"%8.8s", bpb->sFileSystem), 0);
		} else
		{	// FAT32
			BF_BPB			*bpb;

			bpb = (BF_BPB *)p;
			PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", buffer[0],buffer[1],buffer[2]), 0);
			PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", &buffer[3]), 0);
			PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorSize), 0);
			PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", bpb->bAllocUnit, bpb->bAllocUnit, (unsigned)bpb->bAllocUnit * bpb->wSectorSize), 0);
			PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->wReserveSector), 0);
			PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", bpb->bFatNumber), 0);
		//	PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", bpb->wRootEntry), 0);
			PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->wSectorVolume, (INT64)bpb->wSectorSize * bpb->wSectorVolume), 0);
			PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", bpb->bMediaDescriptor), 0);
		//	PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", bpb->wFatSector), 0);
			PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorPerTrack), 0);
			PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", bpb->wHeadNumber), 0);
			PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", bpb->dwHiddenSector), 0);
			PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->dwLongSectorVolume, (INT64)bpb->wSectorSize * bpb->dwLongSectorVolume), 0);
			PutPos( 4, "Big FAT Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->dwBigFatSector), 0);
			PutPos( 4, "Ext Flags"         , COLUMN1+6, Fmt(t1,"%04XH", bpb->wExtFlags), 0);
			PutPos( 4, "FS Version"        , COLUMN1+6, Fmt(t1,"%04XH", bpb->wFsVer), 0);
			PutPos( 4, "Root Cluster"      , COLUMN1  , Fmt(t1,"%10u", bpb->dwRootCluster), 0);
			PutPos( 4, "FS Information"    , COLUMN1  , Fmt(t1,"%10u", bpb->wFsInfo), 0);
			PutPos( 4, "Backup Boot Sector", COLUMN1  , Fmt(t1,"%10u", bpb->wBackupBootSector), 0);
			PutPos( 4, Fmt(t2,"Drive No %s", (bpb->bDrvNo & 0x80) ? "(primary)":""), COLUMN1  , Fmt(t1,"%10u", bpb->bDrvNo & 0x7F), 0);
			PutPos( 4, "Boot Signature"    , COLUMN1+8, Fmt(t1,"%02XH", bpb->bBootSignature), 0);
			PutPos( 4, "Serial No"         , COLUMN1+1, Fmt(t1,"%04X-%04X", bpb->dwSerialNo>>16, bpb->dwSerialNo & 0xFFFF), 0);
			PutPos( 4, "Volume Label"      , COLUMN1-1, Fmt(t1,"%11.11s", bpb->sVolumeLabel), 0);
			PutPos( 4, "File System"       , COLUMN1+2, Fmt(t1,"%8.8s", bpb->sFileSystem), 0);
		}
		
		retcode = 0;
	}
	catch (int ret) { retcode = ret; }
	delete [] buffer;
	return retcode;
}

int		DevTest::PrintNtfsInfo(DevTest::DevParam & devParam, INT64 addr)
{
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;
	BYTE						*buffer = NULL;
	int							retcode;

	if ( devParam.capa == NULL )
		return -1;
	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	blkNum   =           this->devCtl.ScsiGetLBA(devParam.capa->logicalBlockAddress);
	blkLen   = (unsigned)this->devCtl.ScsiGetLBA(devParam.capa->blockLength);
	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	try
	{
		char			t1[256];
		NTFSBootBlock	*ntbb;

		if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr, 1, blkLen, buffer) < (int)blkLen )
			throw -1;
		ntbb = (NTFSBootBlock *)buffer;
		PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", ntbb->jump[0],ntbb->jump[1],ntbb->jump[2]), 0);
		PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", ntbb->oem_id), 0);
		PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorSize), 0);
		PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", ntbb->bAllocUnit, (unsigned)ntbb->bAllocUnit * ntbb->wSectorSize), 0);
		PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", ntbb->wReserveSector), 0);
	//	PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", ntbb->bFatNumber), 0);		// not use
	//	PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", ntbb->wRootEntry), 0);		// not use
	//	PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorVolume), 0);	// not use
		PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", ntbb->bMediaDescriptor), 0);
	//	PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", ntbb->wFatSector), 0);		// not use
		PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorPerTrack), 0);
		PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", ntbb->wHeadNumber), 0);
		PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", ntbb->dwHiddenSector), 0);
	//	PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u", ntbb->dwLongSectorVolume), 0);	// not use
	//	PutPos( 4, "Big FAT Sector"    , COLUMN1  , Fmt(t1,"%10u", ntbb->dwBigFatSector), 0);		// not use
		PutPos( 4, "Total Sectors"     , COLUMN1  , Fmt(t1,"%10"FI64"u (%,I64u bytes)", (INT64)ntbb->totalSectors, (INT64)ntbb->wSectorSize * ntbb->totalSectors), 0);
		PutPos( 4, "MFT Logical Cluster"       , COLUMN1  , Fmt(t1,"%10"FI64"u", ntbb->lcnMFT), 0);
		PutPos( 4, "Mirror MFT Logical Cluster", COLUMN1  , Fmt(t1,"%10"FI64"u", ntbb->lcnMFTMirror), 0);
		PutPos( 4, "File Record Segment", COLUMN1  , Fmt(t1,"%10u", ntbb->dwFileRecSeg), 0);
		PutPos( 4, "Index Block"        , COLUMN1  , Fmt(t1,"%10u", ntbb->dwIndexBlock), 0);
		PutPos( 4, "Volume Serial No"   , COLUMN1-7, Fmt(t1,"%08X-%08X", (unsigned)(ntbb->volumeSerialNo>>I64(32)), (unsigned)ntbb->volumeSerialNo & 0xFFFFFFFF), 0);
		PutPos( 4, "Checksum"           , COLUMN1+2, Fmt(t1,"%08XH", ntbb->dwChecksum), 0);
		retcode = 0;
	}
	catch (int ret) { retcode = ret; }
	delete [] buffer;
	return retcode;
}


int		DevTest::PrintCdrom(DevTest::DevParam & devParam)
{
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;
	DevCtl::SCSITOCDATA			toc;
	char						t1[80], t2[80];
	unsigned					n, tocLength;
	BOOL						msf = FALSE;

	if ( devParam.capa == NULL || devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	blkNum   =           this->devCtl.ScsiGetLBA(devParam.capa->logicalBlockAddress);
	blkLen   = (unsigned)this->devCtl.ScsiGetLBA(devParam.capa->blockLength);

	if ( this->devCtl.ScsiReadToc(pathId, targetId, lun, msf, toc) < 0 )
		return -1;
	tocLength = SC16(toc.tocLength);
	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "### [## ## ##]-[## ## ##] ########## ########## ##:#################### #######
	// "CD-ROM TOC DATA               Start:###       End:###
	// "   [##] ## Data track     ############# ############# Pre-emphasis  DigitCopy
	// "   [##] ## Audio (#ch          ##:##:##    ##########           ON         OK
	fprintf(stdout,"CD/DVD TOC DATA           %13s %13s\n"
	  , Fmt(t1,"Start:%u", toc.startTrack)
	  , Fmt(t2,"End:%u"  , toc.lastTrack )
	);
	fprintf(stdout, "   TrackNo TrackType          StartAddr     TrackSize Pre-emphasis DigitCopy\n");
	for ( n = 0; n < sizeof(toc.track)/sizeof(*toc.track); n++ )
	{
		char		trackNoBuf[40], *trackType, adrBuf[40], timeBuf[40];
		const char	*pe, *dcopy;
		INT64		adr, frame, nextadr;
		unsigned	ff, ss, mm;

		if ( n * sizeof(*toc.track) + 10 >= tocLength )
			break;
		if ( toc.track[n].no == 0xAA )
			strcpy(trackNoBuf, "--");
		else
			Fmt(trackNoBuf, "%2u", toc.track[n].no);
		if ( toc.track[n].adrctl & 0x04 )
		{	// DATA
			trackType = "Data track";
		} else
		{	// Audio
			if ( toc.track[n].adrctl & 0x08 )
				trackType = "Audio (4ch";
			else
				trackType = "Audio (2ch";
		}
		pe    = (toc.track[n].adrctl & 0x01) ? "ON" : "OFF";
		dcopy = (toc.track[n].adrctl & 0x02) ? "OK" : "";
		if ( msf == FALSE )
		{
			adr     = (INT64) SC32(toc.track[n  ].cdAddr);
			nextadr = (INT64) SC32(toc.track[n+1].cdAddr);
		} else
		{
			mm = toc.track[n].cdAddr[1];
			ss = toc.track[n].cdAddr[2];
			ff = toc.track[n].cdAddr[3];
			adr = (INT64)mm * I64(60)*I64(75) + (INT64)ss*I64(75) + (INT64)ff;
		//	sprintf(adrBuf,"%u:%u:%u:%u"
		//	  , toc.track[n].cdAddr[0]
		//	  , toc.track[n].cdAddr[1]
		//	  , toc.track[n].cdAddr[2]
		//	  , toc.track[n].cdAddr[3]
		//	);

			mm = toc.track[n+1].cdAddr[1];
			ss = toc.track[n+1].cdAddr[2];
			ff = toc.track[n+1].cdAddr[3];
			nextadr = (INT64)mm * I64(60)*I64(75) + (INT64)ss * I64(75) + (INT64)ff;
		//	sprintf(timeBuf,"%u:%u:%u:%u"
		//	  , toc.track[n+1].cdAddr[0]
		//	  , toc.track[n+1].cdAddr[1]
		//	  , toc.track[n+1].cdAddr[2]
		//	  , toc.track[n+1].cdAddr[3]
		//	);
		}
		frame = nextadr - adr;
		if ( toc.track[n].adrctl & 0x04 )
		{	// DATA Trsck
			sprintf(adrBuf , "%"FI64"u", (INT64)adr );
			sprintf(timeBuf, "%"FI64"u", (INT64)frame * blkLen);
		} else
		{	// AUDIO Track
			ff = (unsigned)(adr % I64(75));
			ss = (unsigned)((adr/I64(75)) % I64(60));
			mm = (unsigned)((adr/I64(75) / I64(60)));
			sprintf(adrBuf,"%2u:%02u:%02u", mm, ss, ff);

			ff = (unsigned)(frame % I64(75));
			ss = (unsigned)((frame/I64(75)) % I64(60));
			mm = (unsigned)((frame/I64(75) / I64(60)));
			sprintf(timeBuf,"%2u:%02u:%02u", mm, ss, ff);
		}
		// "CD-ROM TOC DATA            Start:###    End:###
		// "   [##] ## Data track     ########## ########## Pre-emphasis DigitCopy
		// "   [##] ## Audio (#ch       ##:##:## ##########           ON        OK
		fprintf(stdout,"   [%2s] %2u %-14s %13s %13s %12s %9s\n"
		  , trackNoBuf
		  , n
		  , trackType
		  , adrBuf
		  , timeBuf
		  , pe
		  , dcopy
		  );
	}
	return 0;
}


//////////////////////////////////////////////////////////////////////////
//	Dump

int		DevTest::PrintDump(DevParam & devParam)
{
	BYTE						*buffer;
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;

	fprintf(stdout,"***** Dump : LBA %"FI64"u *****\n\n", (INT64)this->opt.addr);

	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{	// error
		devParam.capa = NULL;
		memset(&capa, 0, sizeof(capa));
		blkNum = 0;
		blkLen = 0;
	} else
	{
		devParam.capa = &capa;
		blkNum =           this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
		blkLen = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
	}

	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	if ( this->devCtl.ScsiRead(pathId, targetId, lun, (INT64)this->opt.addr, 1, blkLen, buffer) < (int)blkLen )
		return -1;
	this->PrintDumpHex((unsigned)(this->opt.addr * blkLen), blkLen, blkLen, buffer);

	delete [] buffer;
	return 0;
}


//////////////////////////////////////////////////////////////////////////
//	SurfaceScan

int		DevTest::SurfaceScan(DevParam & devParam)
{
	BYTE						*buffer = NULL;
	unsigned					bufferSize, bn, n;
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	INT64						blkNum, addr, errnum = 0;
	unsigned					blkLen, count;
	time_t						tmStart, tmEnd, tmNow, tmLast;
	Clock_t						clkLast, clk;
	BOOL						cmdVerify = TRUE;
	const char					*usecmd = "Verify";

	fprintf(stdout,"***** Surface Scan *****\n\n");

	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{	// error
		devParam.capa = NULL;
		memset(&capa, 0, sizeof(capa));
		blkNum = 0;
		blkLen = 0;
	} else
	{
		devParam.capa = &capa;
		blkNum =           this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
		blkLen = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
	}
	bn         = (this->ben.maxBufferSize + blkLen - 1) / blkLen;
	bufferSize = bn * blkLen;
	if ( this->opt.verify == FALSE || this->devCtl.ScsiVerify(pathId, targetId, lun, I64(0), 1) < 0 )
	{
		usecmd = "Read";
		cmdVerify = FALSE;
		if ( (buffer = new BYTE [bufferSize]) == NULL )
			return -1;
	}
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
	{
		this->InitCpuUsage();
		this->CollectCpuUsage();
	}
	time(&tmStart);
	tmLast = tmStart;
	clkLast = this->Clock();
	if ( (addr = this->opt.addr) >= blkNum )
		addr = blkNum - 1;
	count = 0;
	while ( addr < blkNum )
	{
		time(&tmNow);
		if ( difftime(tmNow, tmLast) > 1 )
		{
			char		speed[80], remain[80], cpuusage[80];
			unsigned	total_sec_per_blk, sec_per_blk, remSec;

			speed[0] = remain[0] = 0;
			clk = this->Clock();
			if ( clk - clkLast > 0 )
			{
				sec_per_blk = (unsigned)( (INT64)CLOCKS_PER_SEC * (INT64)count / (INT64)(clk - clkLast) );
				sprintf(speed, "%u.%1ukB/s"
				  , (unsigned) ( (INT64)sec_per_blk * (INT64)blkLen / I64(1000))
				  , (unsigned) (I64(10) * (INT64)sec_per_blk * (INT64)blkLen / I64(1000)) % 10
				);
				total_sec_per_blk = (unsigned)( (INT64)(addr - this->opt.addr) / (INT64)(difftime(tmNow, tmStart)) );
				if ( sec_per_blk > 0 && total_sec_per_blk > 0 )
				{
					sec_per_blk = (sec_per_blk + total_sec_per_blk) / 2;
					remSec = (unsigned)((blkNum - addr) / (INT64)sec_per_blk);
					this->FormatTimeRemain(remain, remSec);
				}
			}
			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
			{
				this->CollectCpuUsage();
				sprintf(cpuusage, ":CPU%5.1f%%", (double)this->GetCpuUsage() );
			} else
				cpuusage[0] = '\0';
			fprintf(stderr, "%-6s%10"FI64"u/%"FI64"u(%3u.%1u%%)%14s%-9s(Remain%13s\r"
			  , usecmd
			  , (INT64)addr
			  , (INT64)blkNum
			  , (unsigned)(I64(100) * addr / blkNum)
			  , (unsigned)(I64(1000) * addr / blkNum) % 10
			  , speed
			  , cpuusage
			  , remain
			);
			tmLast = tmNow;
			clkLast = clk;
			count = 0;
		}
		{
			BOOL	err;
			int		retry;

			err = FALSE;
			retry = 0;
			n = ( addr + bn < blkNum ) ? bn : (unsigned)(blkNum - addr);
			if ( cmdVerify )
			{
				if ( this->devCtl.ScsiVerify(pathId, targetId, lun, addr, n) < 0 )
					err = TRUE;
			} else
			{
				if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr, n, n * blkLen, buffer) < (int)(n * blkLen) )
					err = TRUE;
			}
			if ( err )
			{
				errnum++;
				retry++;
				fprintf(stdout, "Surface Scan Error! LBA %"FI64"u-%"FI64"u (count %u)\n"
				    , (INT64)addr
				    , (INT64)addr + n - 1
				    , retry
				);
				while ( retry <= this->opt.retry )
				{
					unsigned	m, err;

					err = 0;
					for ( m = 0; m < n; m++ )
					{
						err = 0;
						if ( cmdVerify )
						{
							if ( this->devCtl.ScsiVerify(pathId, targetId, lun, addr + m, 1) < 0 )
								err = TRUE;
						} else
						{
							if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr + m, 1, blkLen, buffer) < (int)blkLen )
								err = TRUE;
						}
						if ( err )
						{
							err++;
							errnum++;
							fprintf(stdout, "Surface Scan Error! LBA %"FI64"u (count %u)\n"
							    , (INT64)addr + m
							    , retry
							);
						}
					}
					if ( err == 0 )
						break;	// OK
					retry++;
				}
			}
			if ( retry > this->opt.retry )
			{	// error
				fprintf(stdout, "Surface Scan Error! LBA %"FI64"u-%"FI64"u\n", (INT64)addr, (INT64)addr + n - 1);
				break;	// erro error exist
			}
		}
		addr += n;
		count += n;
	}
	time(&tmEnd);
	{
		unsigned	sec;
		char		tmbuf[32];

		sec = (unsigned) difftime(tmEnd, tmStart);
		fprintf(stdout, "\nSurface scan completed. %usec (%13s) : Error %"FI64"u \n"
		  , sec, this->FormatTimeRemain(tmbuf, sec)
		  , (INT64)errnum);
	}
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		this->TermCpuUsage();
	if ( buffer != NULL )
		delete [] buffer;
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Read/Write Block

int		DevTest::ReadWriteBlock(DevTest::DevParam & devParam)
{
	BYTE						*buffer = NULL;
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	INT64						blkNum, startAddr, endAddr, maxCount;
	unsigned					blkLen;
	time_t						tmStart, tmEnd, tmNow, tmLast;
	Clock_t						clkLast, clk;
	int							retcode;
	FILE						*fp = NULL;
	const char					*usecmd = this->opt.cmd == DevTest::CMD_READBLOCK ? "Read" : "Write";

	fprintf(stdout,"***** %s Block *****\n\n", usecmd);

	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{	// error
		devParam.capa = NULL;
		memset(&capa, 0, sizeof(capa));
		blkNum = 0;
		blkLen = 0;
	} else
	{
		devParam.capa = &capa;
		blkNum =           this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
		blkLen = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
	}

	if ( (startAddr = this->opt.addr) >= blkNum )
		startAddr = blkNum - 1;
	maxCount = (this->opt.count == I64(0) || this->opt.count > blkNum) ? blkNum : this->opt.count;
	if ( startAddr + maxCount > blkNum )
		maxCount = blkNum - startAddr;
	endAddr = startAddr + maxCount - 1;

	try
	{
		unsigned	bufBlks, bufferSize;
		INT64		addr, lbacount;
		unsigned	count;

		bufBlks    = (this->ben.maxBufferSize + blkLen - 1) / blkLen;
		bufferSize = bufBlks * blkLen;
		if ( (buffer = new BYTE [bufferSize]) == NULL )
			return -1;

		fprintf(stdout,"  Start Address %10"FI64"u\n", (INT64)startAddr);
		fprintf(stdout,"  %-5s Count   %10"FI64"u\n", usecmd, (INT64)maxCount  );
		fprintf(stdout,"  Buffer size   %10u\n", (unsigned)bufferSize);

		if ( this->opt.cmd == DevTest::CMD_READBLOCK )
		{	// disk to file(out)
			if ( this->opt.filename != NULL )
			{
				if ( (fp = fopen(this->opt.filename, "wb")) == NULL )
				{
					fprintf(stderr, "File open error! (%s)\n", this->opt.filename);
					throw -1;
				}
			} else
				fp = stdout;
		} else
		{	// file(in) to disk
			if ( this->opt.zeroFill )
			{
				fp = NULL;
				memset(buffer, 0, blkLen);
			} else
			{
				if ( this->opt.filename != NULL )
				{
					if ( (fp = fopen(this->opt.filename, "rb")) == NULL )
					{
						fprintf(stderr, "File open error! (%s)\n", this->opt.filename);
						throw -1;
					}
				} else
					fp = stdin;
			}
		}

		if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
		{
			fprintf( this->ben.csvResult, "time(sec), addr(LBA), addr(bytes), blocks/s, kB/s\n");
		}
		if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		{
			this->InitCpuUsage();
			this->CollectCpuUsage();
		}
		time(&tmStart);
		tmLast = tmStart;
		clkLast = this->Clock();
		count = 0;
		addr = startAddr;
		lbacount = I64(0);
		for (;;)
		{
			time(&tmNow);
			if ( difftime(tmNow, tmLast) > 1 || lbacount == maxCount )
			{
				char		speed[80], remain[80], cpuusage[80];
				unsigned	total_sec_per_blk, sec_per_blk, remSec;

				speed[0] = remain[0] = 0;
				clk = this->Clock();
				if ( clk - clkLast > 0 )
				{
					sec_per_blk = (unsigned)( (INT64)CLOCKS_PER_SEC * (INT64)count / (INT64)(clk - clkLast) );
					sprintf(speed, "%u.%1u"
					  , (unsigned) ( (INT64)sec_per_blk * (INT64)blkLen / I64(1000))
					  , (unsigned) (I64(10) * (INT64)sec_per_blk * (INT64)blkLen / I64(1000)) % 10
					);
					total_sec_per_blk = (unsigned)( (INT64)(lbacount) / (INT64)(difftime(tmNow, tmStart)) );
					if ( sec_per_blk > 0 && total_sec_per_blk > 0 )
					{
						sec_per_blk = (sec_per_blk + total_sec_per_blk) / 2;
						remSec = (unsigned)((maxCount - lbacount) / (INT64)sec_per_blk);
						this->FormatTimeRemain(remain, remSec);
					}
				} else
				{
					strcpy(speed, "?.?");
					strcpy(remain, "????");
				}
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				{
					this->CollectCpuUsage();
					sprintf(cpuusage, ":CPU%5.1f%%", (double)this->GetCpuUsage() );
				} else
					cpuusage[0] = '\0';
				fprintf(stderr, "%-6s%10"FI64"u/%"FI64"u(%3u.%1u%%)%10skB/s%-9s(Remain%13s\r"
				  , usecmd
				  , (INT64)addr
				  , (INT64)endAddr
				  , (unsigned)(I64(100)  * (lbacount) / maxCount)
				  , (unsigned)(I64(1000) * (lbacount) / maxCount) % 10
				  , speed
				  , cpuusage
				  , remain
				);
				if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
				{
					fprintf( this->ben.csvResult, "%u, %"FI64"u, %"FI64"u, %u, %s \n" 
					    , (unsigned)difftime(tmNow, tmStart)
					    , (INT64)addr
					    , (INT64)addr * blkLen
					    , sec_per_blk
					    , speed
					);
				}
				tmLast = tmNow;
				clkLast = clk;
				count = 0;
			}
			if ( lbacount >= maxCount )
				break;

			{
				unsigned	n;

				if ( lbacount + (INT64)bufBlks <= maxCount )
					n = bufBlks;
				else
					n = (unsigned)(maxCount - lbacount);
				if ( n == 0 )
					break;	// error
				if ( this->opt.cmd == DevTest::CMD_READBLOCK )
				{
					// Read
					if ( this->devCtl.ScsiRead(pathId, targetId, lun, addr, n, n * blkLen, buffer) < (int)(n * blkLen) )
					{
						fprintf(stderr, "ScsiRead Error! (lba %"FI64"u)\n", (INT64)addr);
						throw -1;
					}
					if ( fwrite(buffer, 1, n * blkLen, fp) < n * blkLen )
					{
						fprintf(stderr, "File Write Error!\n");
						throw -1;
					}
				} else
				{
					// Write
					if ( this->opt.zeroFill )
					{
					} else
					{
						if ( fread(buffer, 1, n * blkLen, fp) < n * blkLen )
						{
							fprintf(stderr, "File Read Error!\n");
							throw -1;
						}
					}
					// Delete Source Code (ScsiWrite)
				}
				lbacount += n;
				addr += n;
				count += n;
			}
		}
		time(&tmEnd);

		{
			unsigned	sec;
			char		tmbuf[32];

			sec = (unsigned) difftime(tmEnd, tmStart);
			fprintf(stdout, "\n%s completed. %usec (%13s) \n"
			  , usecmd
			  , sec, this->FormatTimeRemain(tmbuf, sec)
			);
		}
		retcode = 0;
	} catch (int code) { retcode = code; }

	if ( fp != NULL && fp != stdin && fp != stdout )
		fclose(fp);
	if ( buffer != NULL )
		delete [] buffer;
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		this->TermCpuUsage();

	return retcode;
}
