/////////////////////////////////////////////////////////////////////////
//	"devtest.h"
//------------------------------------------------------------------------
//	$Id: devtest.h 127 2004-07-12 10:49:17Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	DEVTEST_H_INCLUDED
#define	DEVTEST_H_INCLUDED

#define	DEVTEST_USE_CDROM		// Scan "Cdrom#" Device
#define	DEVTEST_USE_UDFLIB		// Check UDF Disc
//#define	DEVTEST_USE_STL_LIST
#define	DEVTEST_DUMP_CSTYLE		// for Debug & Development

#if		defined(WIN32)
# include	"win32/devctl.h"
# include	"perflib.h"
#elif	defined(LINUX)
# include	"linux/devctl.h"
#endif

#if	defined(DEVTEST_USE_UDFLIB)
# include	"udflib.h"
#endif

#if	defined(DEVTEST_USE_STL_LIST)
# include	<list>
#endif

#include	<stdio.h>

#if	defined(WIN32)
# define	DEVTEST_TITLE	"***** DevTest - Device Benchmark Test Program for Windows 2000/XP *****"
#elif defined(LINUX)
# define	DEVTEST_TITLE	"***** DevTest - Device Benchmark Test Program for Linux (beta)    *****"
#endif
#define		DEVTEST_VERSION	"                                        Version 1.03a edition #01"

class	DevTest
{
public:
	enum devtest_enum
	{
		SECTOR_SIZE_MAX    = 65536,
		MAX_BUFFER_SIZE    = 65536,
		MIN_BUFFER_SIZE    = 256,
		TEST_TIME_SEC      = 3,
		RND_TEST_TIME_SEC  = 10,
		RANDOM_SCALE       = 32768,
		TRACK_POS_SCALE    = 1024,
		RWRATIO_SCALE      = 1000,
		COMMAND_TEST_COUNT = 128,
		IBM_DISK_BLOCKSIZE = 512,

		COLUMN1            = 38,
		COLUMN2            = 49,
		COLUMN3            = 60,

		CD_SPEED           =  150000,
		DVD_SPEED          = 1385000,

		MODE_CDB6          =  6,
		MODE_CDB10         = 10,
	};
	enum byteunit_enum
	{
		BYTEUNIT_210,	// 2^10
		BYTEUNIT_103,	// 10^3
	};
	enum benchflag_enum
	{
		BEN_READ        = 0x00000001,
		BEN_WRITE       = 0x00000002,
		BEN_RW          = BEN_READ | BEN_WRITE,
		BEN_RWMASK      = BEN_READ | BEN_WRITE,

		BEN_SEQUENTIAL  = 0x00000010,
		BEN_RANDOM      = 0x00000020,

		BEN_TRACK_START = 0x00000100,
		BEN_TRACK_END   = 0x00000200,
		BEN_TRACK_USER  = 0x00000400,

		BEN_BLOCK_MIN   = 0x00001000,
		BEN_BLOCK_MAX   = 0x00002000,
		BEN_BLOCK_USER  = 0x00004000,
	};
	enum cpuusage_enum
	{
		CPUUSAGE_OFF  ,
		CPUUSAGE_SELF ,
#if	defined(WIN32)
		CPUUSAGE_ALL  ,
#endif
#if	defined(LINUX)
		CPUUSAGE_SYS  ,
		CPUUSAGE_USER ,
#endif
	};
	enum command_enum
	{
		CMD_SCAN,
		CMD_SHOW,
		CMD_DUMP,
		CMD_BENCHMARK,
		CMD_SURFSCAN,
		CMD_READBLOCK ,
	};
#if	defined(WIN32)
	enum scanmode_enum
	{
		SCAN_OFF  ,
		SCAN_DRIVE,
		SCAN_DISK ,
		SCAN_SCSI ,
#if	defined(DEVTEST_USE_CDROM)
		SCAN_CDROM,
#endif
		SCAN_ALL  ,
	};
#endif

	typedef INT64				Clock_t;

	struct DevParam
	{
		char						*devname;
		char						*desc;
		DevCtl::ScsiInqData	 		*inqData;
		DevCtl::SCSICAPACITYDATA	*capa;
		int							pathId;
		int							targetId;
		int							lun;
		DevParam();
		virtual ~DevParam();
		int					SetDevName(const char * text);
		int					SetDesc(const char * text);
		char *				DevParam::GetDesc(unsigned max = 0);
	};
#if	defined(DEVTEST_USE_STL_LIST)
	typedef std::list<DevTest::DevParam *>	DevArgList;
#else

	struct DevTargetParam
	{
#if	defined(WIN32)
		DevCtl::ScsiDevData 	*scsiDev;
		DevCtl::ScsiBusData 	*busData;
#endif
		DevCtl::ScsiInqData 	*inqData;
		DevCtl::ScsiInqData		inqDataWork;
		BOOL					setInqData;
	};

	class	DevArgList
	{
	public:
		class	Item
		{
		public:
			DevParam	*p;
			Item		*next, *prev;
		};
		class	iterator
		{
		friend class	DevArgList;
		private:
			Item			*item;
		public:
			iterator() { this->item = NULL; }

			iterator &	operator = (const iterator & x) { this->item = x.item; return *this; }
			iterator &	operator = (const void * p) { this->item = (Item *)p; return *this; }
			DevParam * 	operator * () const { return this->item->p; }
			DevParam **	operator -> () const { return &(this->item->p); }
			int			operator == (const void * ptr) const { return this->item == ptr; }
			int			operator != (const void * ptr) const { return this->item != ptr; }
			int			operator == (const iterator & it) const { return this->item == it.item; }
			int			operator != (const iterator & it) const { return this->item != it.item; }
			iterator &	operator ++ () { if ( this->item != NULL ) this->item = this->item->next; return *this; }
			iterator &	operator -- () { if ( this->item != NULL ) this->item = this->item->prev; return *this; }
			iterator	operator ++ (int) { iterator it = *this; ++(*this); return it; }
			iterator	operator -- (int) { iterator it = *this; --(*this); return it; }
		};
	private:
		unsigned		numItem;
		iterator		itop, ilast;
		iterator		iend, icur;
	public:
		DevArgList();
		~DevArgList();

		void		clear(void);
		iterator 	insert(iterator & base, const DevParam * s);
		iterator 	push_back(const DevParam * x) { iterator base; base.item = NULL; return this->insert(base, x); }
		void		remove(iterator & x);
		iterator 	begin(void) const { return this->itop; }
		iterator 	end(void) const { return this->iend; }
		unsigned	size(void) const { return this->numItem; }
	};
#endif
	typedef DevArgList::iterator			DevArgIterator;

	struct tagOption
	{
		int			cmd;
		int			scanMode;
		unsigned	byteUnit;
		BOOL		printPartition;
		BOOL		printDevInfo;
		BOOL		printDetail;
		BOOL		dumpModeSense;
		BOOL		benchCommand;
		BOOL		benchSeek;
		BOOL		benchRead;
		BOOL		benchWrite;
		BOOL		benchRandomRW;
		unsigned	benchRwRatio;
		const char *benchCsvResult;
		INT64		addr;
		INT64		count;
		INT64		readAddr, writeAddr;
		BOOL		zeroFill;
		int			retry;
		BOOL		verify;
		unsigned	defBlockSize;
		unsigned	mode_cdb;
#if	defined(DEVTEST_DUMP_CSTYLE)
		BOOL		dumpCStyle;
#endif
		const char	*filename;
		unsigned	cpuUsage;
	} opt;

	DevCtl		devCtl;
	struct tagBen
	{
		unsigned	mode;
		unsigned	maxBufferSize;
		unsigned	testTimeSec;
		unsigned	randomRWTimeSec;
		unsigned	trackPos;		// 0 - TRACK_POS_SCALE-1
		unsigned	blockSize;
		unsigned	seekSector;		// sector/track for Seek test
		INT64		discSpeed;
#if	defined(WIN32)
		BOOL		useWinFile;
#endif
		FILE		*csvResult;
	} ben;
	DevParam		*srcDevParam;
	DevParam		*dstDevParam;


#if	defined(WIN32)
	PerfLib				perflib;
	PerfLib::Counter	counter;
#endif

public:
	DevTest();
	~DevTest();

	int		ScanDevice(DevParam & devParam);
	int		ScanDeviceDrive(DevParam & devParam);
	int		ScanDeviceDisk(DevParam & devParam);
	int		ScanDeviceScsi(DevParam & devParam);
#if	defined(DEVTEST_USE_CDROM)
	int		ScanDeviceCdrom(DevParam & devParam);
#endif
	int		ScanDeviceCheck(DevParam & devParam);
	int		OpenTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam);
	void	CloseTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam);
	int		TargetDevice(DevTest::DevParam & devParam);
	int		PrintDevice(DevParam & devParam);
	int		PrintPartition(DevParam & devParam);
	int		PrintIbmPartition(DevParam & devParam, int & ext, INT64 & extAddr, INT64 baseAddr, BYTE * buffer);
	int		PrintFatInfo(DevTest::DevParam & devParam, INT64 addr);
	int		PrintNtfsInfo(DevTest::DevParam & devParam, INT64 addr);
	int		PrintCdrom(DevParam & devParam);
	int		PrintDump(DevParam & devParam);
	int		SurfaceScan(DevParam & devParam);
	int		ReadWriteBlock(DevTest::DevParam & devParam);

	int		PrintDevInfo(DevParam & devParam);
	int		PrintDevInfo_vpd(DevParam & devParam);
	int		PrintDevInfo_logSense(DevTest::DevParam & devParam);
	int		PrintDevInfo_modeSense(DevParam & devParam);
	int		PrintPagecode2A(DevTest::DevParam & devParam, unsigned size, void * buffer);
	int		PrintDevInfo_defectData(DevTest::DevParam & devParam);
	int		PrintDevInfo_getConfiguration(DevTest::DevParam & devParam);
	static char *	GetFeatureCodeName(char * buffer, unsigned fcode);
	static char *	GetProfileDefinitions(char * buffer, unsigned profileNumber);
	int		PrintDevInfo_dvdStructure(DevTest::DevParam & devParam);
	int		DumpParam(unsigned addr, unsigned off, unsigned bufsiz, BYTE * buffer);

	int		BlockCopy(DevTest::DevParam & srcDev, DevTest::DevParam & dstParam);

	int		BenchReadWrite(DevParam & devParam, unsigned benmode);
	int		BenchRandomReadWrite(DevTest::DevParam & devParam, unsigned rwratio);
	int		BenchCommand(DevParam & devParam, unsigned benmode);
	int		BenchSeek(DevParam & devParam, unsigned benmode);
	int		BenchRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer);

	static unsigned	SC32(const BYTE * p) { return DevCtl::ScsiUint32(p); }
	static unsigned	SC24(const BYTE * p) { return DevCtl::ScsiUint24(p); }
	static unsigned	SC16(const BYTE * p) { return DevCtl::ScsiUint16(p); }
	static const char *	GetOnOff(BOOL flag) { return flag ? "ON" : "OFF"; }
	static int		fprintf(FILE * fp, const char * form, ... );
	static char *	Fmt(char * buffer, const char * form, ... );
	static void		PutFmt(int spc, const char * text1, const char * val, const char *text2);
	static void		PutPos(int pos, ... );
	static int		GetNonZeroSize(unsigned size, BYTE * buffer);
	void			PrintDumpOct(const char * msg, int ascii, unsigned size, BYTE * buffer);
	void			PrintDumpHex(unsigned addr, unsigned blockSize, unsigned size, BYTE * buffer);
#if	defined(DEVTEST_DUMP_CSTYLE)
	void			PrintDumpCStyle(unsigned size, BYTE * buffer);
#endif
	char *			FormatSize(char * buffer, INT64 size);
	char *			FormatTimeRemain(char * buffer, unsigned sec);
	char *			FormatTime(char * buffer, unsigned sec);
#if	defined(WIN32)
	static char *	GetErrMsg(int errcode);
#endif
	char *			GetPartitionType(unsigned type);
	Clock_t			Clock(void);

#if	defined(DEVTEST_USE_UDFLIB)
	//////////////////////////////////////////////////////////////////////
	//	UDF
public:
	UdfLib				udflib;

	int			PrintDisc(DevParam & devParam);
#endif

	//////////////////////////////////////////////////////////////////////
	//	Linux ext2
public:
	int			PrintExt2Info(DevParam & devParam, INT64 addr, INT64 addrMax);

	//////////////////////////////////////////////////////////////////////
	//	CPU USAGE
public:
	int			InitCpuUsage(void);
	void		CollectCpuUsage(void);
	double		GetCpuUsage(void);
	void		TermCpuUsage(void);
};

#if	defined(LINUX)
extern int		strcmpi(const char *s1, const char *s2);
extern INT64	_atoi64(const char *s);
#endif

#endif
