/////////////////////////////////////////////////////////////////////////
//	"devtest_bench.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_bench.cpp 126 2004-07-12 10:39:31Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<time.h>
#include	"devtest.h"
#include	"defs.h"

#define	GET_BEN_RWMODE(mode_)	((mode_) & DevTest::BEN_RWMASK)
#define	IS_BENREAD(mode_)		(GET_BEN_RWMODE(mode_) == DevTest::BEN_READ)
#define	IS_BENWRITE(mode_)		(GET_BEN_RWMODE(mode_) == DevTest::BEN_WRITE)

int		DevTest::BenchReadWrite(DevTest::DevParam & devParam, unsigned benmode)
{
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	BYTE 						*buffer = NULL;
	int							retcode;
	enum test_enum
	{
		TEST_SEQ,
		TEST_RND,
		TEST_NUM,
	};
	enum trkmd_enum
	{
		TRKMD_START,
		TRKMD_USER ,
		TRKMD_END  ,
		TRKMD_AVE  ,
		TRKMD_NUM  ,
	};
	enum blkmd_enum
	{
		BLKMD_MIN ,
		BLKMD_USER,
		BLKMD_MAX ,
		BLKMD_NUM ,
	};
	static char * TestText[TEST_NUM] =
	{
		"Sequential",
		"Random"    ,
	};
	unsigned	sectorSize;
	INT64		maxblk;

	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{
		fprintf(stderr,"ScsiReadCapacity error!\n");
		return -1;
	}
	devParam.capa = &capa;
	maxblk     = this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
	sectorSize = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);

	fprintf(stdout, "\n");
	try
	{
		int			test;

		if ( (buffer = new BYTE [this->ben.maxBufferSize]) == NULL )
		{
			fprintf(stderr,"Out of Memory!\n");
			throw -1;
		}
		if ( IS_BENWRITE(benmode) )
		{
			unsigned	n;

			for ( n = 0; n < this->ben.maxBufferSize; n++ )
				buffer[n] = (BYTE)rand();
		}

		fprintf(stdout, "%-16s /%7s %7s : %13s (%13s)%s%s\n"
		  , "TEST MODE"
		  , "TRACK"
		  , "SIZE"
		  , this->opt.byteUnit == BYTEUNIT_210 ? "Speed KB/s" : "Speed kB/s"
		  , "Speed Mbps"
		  , this->ben.discSpeed == 0           ? "" : " CD/DVD"
		  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : "  CPU %"
		);

		///////////////////////////////////////////////////////////////
		//	SEQUENTIAL or RANDOM
		for ( test = 0; test < TEST_NUM; test++ )
		{
			int			trkmd, trackTestCount;
			INT64		restbl[TRKMD_NUM][BLKMD_NUM];

			if ( test == TEST_SEQ && (benmode & BEN_SEQUENTIAL) == 0 )
				continue;	// skip
			if ( test == TEST_RND && (benmode & BEN_RANDOM    ) == 0 )
				continue;	// skip
			trackTestCount = 0;

			///////////////////////////////////////////////////////////////
			//	TRACK ZONE
			for ( trkmd = 0; trkmd < TRKMD_NUM; trkmd++)
			{
				int			blkmd;
				char		*textTrack, textTrackBuf[80];
				INT64		startPos;
				unsigned	blkSizeMem[BLKMD_NUM];

				memset(blkSizeMem, 0, sizeof(blkSizeMem));
				if ( test == TEST_SEQ )
				{
					if ( trkmd == TRKMD_START && (benmode & BEN_TRACK_START) == 0 )
						continue;
					if ( trkmd == TRKMD_USER  && (benmode & BEN_TRACK_USER ) == 0 )
						continue;
					if ( trkmd == TRKMD_END   && (benmode & BEN_TRACK_END  ) == 0 )
						continue;
					if ( trkmd == TRKMD_AVE   && trackTestCount == 0 )
						continue;
					switch ( trkmd )
					{
						case TRKMD_START:
							startPos = 0;
							textTrack = "START";
							break;
						case TRKMD_USER:
							startPos = maxblk * (INT64)this->ben.trackPos / (INT64)TRACK_POS_SCALE;
							sprintf(textTrackBuf,"%3u%%", 100 * this->ben.trackPos / TRACK_POS_SCALE);
							textTrack = textTrackBuf;
							break;
						case TRKMD_END:
							startPos = maxblk - (I64(5) * maxblk / I64(100));	// 5% in
							textTrack = "END";
							break;
						case TRKMD_AVE:
							textTrack = "AVERAGE";
							break;
					}
				} else
				{	// TEST_RAN
					startPos = 0;
					textTrack = "";
					if ( trkmd != TRKMD_START )
						continue;
				}
				fprintf(stdout, "---------------- -------- ------- : ------------- ---------------%s%s\n"
				  , this->ben.discSpeed == 0           ? "" : " ------"
				  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " ------"
				);

				//////////////////////////////////////////////////////////
				//	BLOCK SIZE
				for ( blkmd = 0; blkmd < BLKMD_NUM; blkmd++ )
				{
					unsigned	count, blkCount, blkSize;
					int			rwcount;
					INT64		pos, total_count;
					Clock_t		clk, clklast, clkstart, clkend;
					double		cpuUsage;

					if ( blkmd == BLKMD_MIN  && (benmode & BEN_BLOCK_MIN ) == 0 )
						continue;
					if ( blkmd == BLKMD_USER && (benmode & BEN_BLOCK_USER) == 0 )
						continue;
					if ( blkmd == BLKMD_MAX  && (benmode & BEN_BLOCK_MAX ) == 0 )
						continue;

					switch ( blkmd )
					{
						case BLKMD_MIN:
							blkCount = 1;
							blkSize =  blkCount * sectorSize;
							break;
						case BLKMD_MAX:
							blkCount = this->ben.maxBufferSize / sectorSize;
							blkSize =  blkCount * sectorSize;
							break;
						case BLKMD_USER:
							blkSize =  ((this->ben.blockSize + sectorSize-1)/sectorSize) * sectorSize;
							blkCount = blkSize / sectorSize;
							break;
					}
					if ( blkCount == 0 || blkSize > this->ben.maxBufferSize )
					{
						fprintf(stderr, "Blocksize (count:%u  size:%u maxBufferSize:%u) error!\n"
						  , blkCount
						  , blkSize
						  , this->ben.maxBufferSize
						);
						continue;
					}
					{
						int			i, flag;

						flag = 0;
						for ( i = 0; i < BLKMD_NUM; i++ )
						{
							if ( blkSizeMem[i] == blkSize )
							{
								flag = 1;
								break;
							}
						}
						if ( flag )
							continue;
						blkSizeMem[blkmd] = blkSize;
					}

					//////////////////////////////////////////////////////
					//	PRINT TEST TEXT
					{
						const char	*rwmsg;

						switch ( GET_BEN_RWMODE(benmode) )
						{
							case BEN_READ : rwmsg = "Read" ; break;
							case BEN_WRITE: rwmsg = "Write"; break;
						}
						fprintf(stdout, "%-10s %-5s /%7s %6uB : "
						  , TestText[test]
						  , rwmsg
						  , textTrack
						  , blkSize
						);
					}

					if ( trkmd != TRKMD_AVE )
					{
						MSLEEP(500);
						srand(0);

						if ( this->opt.cpuUsage != CPUUSAGE_OFF )
							this->CollectCpuUsage();

						pos = startPos;
						total_count = 0;
						count = blkCount;
						clkstart = clklast = this->Clock();
						clkend = clkstart + this->ben.testTimeSec * CLOCKS_PER_SEC;
						while ( (clk = this->Clock()) < clkend )
						{
							if ( clk < clklast )
							{
								fprintf(stderr,"Clock error!\n");
								throw -1;	// clock error!
							}
							clklast = clk;
							if ( pos + (INT64)count >= maxblk )
								pos = startPos;
							if ( IS_BENREAD(benmode) )
							{
								if ( (rwcount = this->BenchRead(pathId, targetId, lun, (INT64)pos, count, blkSize, buffer)) < (int)blkSize )
								{
#if		defined(WIN32)
									DWORD	err;

									err = ::GetLastError();
									fprintf(stderr, "ScsiRead error!: %s\n", this->GetErrMsg(err));
									if ( err == ERROR_INVALID_PARAMETER && blkCount == this->ben.maxBufferSize/sectorSize )
										fprintf(stderr, "Try '--maxBufferSize' or '--winfile' option.\n");
#elif	defined(LINUX)
									fprintf(stderr, "ScsiRead error!\n");
#endif
									throw -1;	// error
								}
								rwcount = rwcount / sectorSize;
							// Delete Source Code: Write Benchmark
							}
							total_count += (INT64)rwcount;
							switch ( test )
							{
								case TEST_SEQ:
									pos += (INT64)rwcount;
									if ( pos >= maxblk )
										pos = startPos;
									break;
								case TEST_RND:
									do
									{
										pos = maxblk * (INT64)(rand() % RANDOM_SCALE) / (INT64)RANDOM_SCALE;
									} while ( pos > (maxblk - count) );
									break;
							}
						} // end of while()
						clkend = this->Clock();
						if ( this->opt.cpuUsage != CPUUSAGE_OFF )
						{
							this->CollectCpuUsage();
							cpuUsage = this->GetCpuUsage();
						}
					}

					// result
					{
						INT64		speed;	// byte / sec
						BOOL		disp;

						disp = FALSE;
						if ( trkmd != TRKMD_AVE )
						{
							clk = clkend - clkstart;
							speed = (INT64)CLOCKS_PER_SEC * total_count * (INT64) sectorSize / (INT64)clk;
							disp = TRUE;
						} else
						{
							if ( trackTestCount > 1 )
							{
								int			n;

								speed = 0;
								for ( n = 0; n < trackTestCount; n++ )
									speed += restbl[n][blkmd];
								speed /= trackTestCount;
								disp = TRUE;
							}
						}
						if ( disp )
						{
							char	kbs_n[80], kbs[80], mbps[80], discspeed[16],cpuusage[16];
							int		kiro;
							INT64	ub;

							switch ( this->opt.byteUnit )
							{
								case BYTEUNIT_210:
									kiro = 'K';
									ub = I64(1024);
									break;
								case BYTEUNIT_103:
								default:
									kiro = 'k';
									ub = I64(1000);
									break;
							}
							sprintf(kbs_n, "%u.%1u"
							  , (unsigned)(speed / ub)
							  , (unsigned)((I64(10) * speed / ub) % I64(10))
							);
							sprintf(kbs, "%s %cB/s", kbs_n, kiro);
							sprintf(mbps
							  , "%u.%02u Mbps"
							  , (unsigned)(speed * I64(8)/ (ub*ub))
							  , (unsigned)((I64(100) * speed * I64(8) / (ub*ub)) % I64(100))
							);
							if ( this->ben.discSpeed > 0 )
							{
								char	t1[80];

								sprintf(t1
								  , "x%u.%1u"
								  , (unsigned)(I64(1000) * speed / this->ben.discSpeed / I64(1000))
								  , (unsigned)((I64(10) * I64(1000) * speed / this->ben.discSpeed / I64(1000)) % I64(10))
								);
								sprintf(discspeed, " %6s", t1);
							} else
								discspeed[0] = '\0';
							if ( this->opt.cpuUsage != CPUUSAGE_OFF )
							{
								if ( trkmd != TRKMD_AVE )
									sprintf(cpuusage, " %5.1f%%", cpuUsage);
								else
									strcpy(cpuusage, "       ");
							} else
								cpuusage[0] = '\0';
							fprintf(stdout, "%13s (%13s)%s%s\n"
							  , kbs
							  , mbps
							  , discspeed
							  , cpuusage
							);
							if ( this->opt.benchCsvResult != NULL )
							{
								fprintf(this->ben.csvResult, "%-10s %-5s /%7s %6uB , %s\n"
								  , TestText[test]
								  , (benmode & BEN_READ) != 0 ? "Read" : "Write"
								  , textTrack
								  , blkSize
								  , kbs_n
								);
							}
						}
						restbl[trackTestCount][blkmd] = speed;
					}
				} // end of for(blkmd;;)
			//	fprintf(stdout, "\n");
				++trackTestCount;
			} // end of for(trkmd;;)
		} // end of for(test;;)
		retcode = 0;
	}
	catch (int code) { retcode = code; }
	if ( buffer != NULL )
		delete [] buffer;
	return retcode;
}

#define	BLKTBL_MAX	(64)

int		DevTest::BenchRandomReadWrite(DevTest::DevParam & devParam, unsigned rwratio)
{
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	BYTE 						*rbuf = NULL, *wbuf = NULL;
	int							retcode;
	unsigned					sectorSize, maxRwBlk;
	INT64						maxblk;
	typedef struct
	{
		unsigned	blks;
	} BlkItem;
	BlkItem						*rtbl = NULL, *wtbl = NULL;
	unsigned					tblnum = 32;

	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{
		fprintf(stderr,"ScsiReadCapacity error!\n");
		return -1;
	}
	devParam.capa = &capa;
	maxblk     = this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
	sectorSize = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
	maxRwBlk   = this->ben.maxBufferSize / sectorSize;

	fprintf(stdout, "\n");
	try
	{
		int			test;
		enum test_enum
		{
			TEST_RW ,
			TEST_R  ,
			TEST_W  ,
			TEST_MAX,
		};

		rbuf = new BYTE [this->ben.maxBufferSize];
		wbuf = new BYTE [this->ben.maxBufferSize];
		rtbl = new BlkItem [tblnum];
		wtbl = new BlkItem [tblnum];
		if ( rbuf == NULL || wbuf == NULL || rtbl == NULL || wtbl == NULL )
		{
			fprintf(stderr,"Out of Memory!\n");
			throw -1;
		}
		{
			unsigned	n;

			for ( n = 0; n < this->ben.maxBufferSize; n++ )
				wbuf[n] = (BYTE) rand();
			for ( n = 0; n < tblnum; n++ )
			{
				if ( n == tblnum - 1 )
					rtbl[n].blks = wtbl[n].blks = maxRwBlk;
				else
					rtbl[n].blks = wtbl[n].blks = (n * maxRwBlk / tblnum) + 1;
			}
		}

		for ( test = 0; test < TEST_MAX; test++ )
		{
			unsigned	rcount, wcount;
			INT64		rsize, wsize;
			unsigned	ratio;
			Clock_t		clk, clklast, clkstart, clkend;
			double		cpuUsage;
			char		testMsgBuf[80];

			//////////////////////////////////////////////////////////////
			// Guide
			//                                                            Mbps   CD/DVD  CPU %
			//   ---------------- ----------- ---- : ------------- (-------------) ------ ------ */
			//   Random R/W 50:50 Read  Speed 55.5           kB/s
			//   Random R/W 50:50 Write Speed                kB/s
			//   Random R/W 50:50 Read  Count                io/s
			//   Random R/W 50:50 Write Count                io/s
			fprintf(stdout
			  , "---------------- ----------- ---- : ------------- (-------------)%s%s\n"
			  , this->ben.discSpeed == 0           ? "" : " ------"
			  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " ------"
			);

			srand(0);
			switch ( test )
			{
				case TEST_RW: ratio = RWRATIO_SCALE / 2      ; break;
				case TEST_R : ratio = rwratio                ; break;
				case TEST_W : ratio = RWRATIO_SCALE - rwratio; break;
			}
			if ( test != TEST_RW && ratio == RWRATIO_SCALE / 2 )
				continue;
			sprintf(testMsgBuf, "Random R/W %02u:%02u"
			  , (100 *                  ratio  / RWRATIO_SCALE)
			  , (100 * (RWRATIO_SCALE - ratio) / RWRATIO_SCALE)
			);
			{
				unsigned	n, n1, n2;
				BlkItem		bi;

				for ( n = 0; n < tblnum; n++ )
				{
					n1 = rand() % tblnum;
					n2 = rand() % tblnum;
					bi.blks = rtbl[n1].blks;
					rtbl[n1].blks = rtbl[n2].blks;
					rtbl[n2].blks = bi.blks;

					n1 = rand() % tblnum;
					n2 = rand() % tblnum;
					bi.blks = wtbl[n1].blks;
					wtbl[n1].blks = wtbl[n2].blks;
					wtbl[n2].blks = bi.blks;
				}
			}

			rcount = 0;
			wcount = 0;
			rsize = 0;
			wsize = 0;

			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				this->CollectCpuUsage();

			clkstart = clklast = this->Clock();
			clkend = clkstart + this->ben.randomRWTimeSec * CLOCKS_PER_SEC;
			while ( (clk = this->Clock()) < clkend )
			{
				INT64		pos;
				int			rescount;
				unsigned	rwcount, rwsize;
				int			rw_mode;

				if ( clk < clklast )
				{
					fprintf(stderr,"Clock error!\n");
					throw -1;	// clock error!
				}
				clklast = clk;
				rw_mode = (unsigned)(rand() % RWRATIO_SCALE) < ratio ? 0 : 1;
				pos = maxblk * (INT64)(rand() % RANDOM_SCALE) / (INT64)RANDOM_SCALE;
				rwcount = rw_mode == 0 ? rtbl[rcount % tblnum].blks : wtbl[wcount % tblnum].blks;
				if ( pos + (INT64)rwcount >= maxblk )
					pos = maxblk - (INT64) rwcount;
				rwsize = rwcount * sectorSize;

				if ( rw_mode == 0 )
				{
					if ( (rescount = this->BenchRead(pathId, targetId, lun, (INT64)pos, rwcount, rwsize, rbuf)) < (int)rwsize )
					{
#if		defined(WIN32)
						DWORD	err;

						err = ::GetLastError();
						fprintf(stderr, "ScsiRead error!: %s\n", this->GetErrMsg(err));
#elif	defined(LINUX)
						fprintf(stderr, "ScsiRead error!\n");
#endif
						fprintf(stderr, "result %d, LBA %"FI64"u, count %u, size %u\n", (int)rescount, pos, rwcount, rwsize);
						throw -1;	// error
					}
					rcount++;
					rsize += (INT64)rwsize;
				} else
				{
					// Delete Source Code: Write Benchmark
					wcount++;
					wsize += (INT64)rwsize;
				}
			} // end of while (clk)
			clkend = this->Clock();
			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
			{
				this->CollectCpuUsage();
				cpuUsage = this->GetCpuUsage();
			}

			//////////////////////////////////////////////////////////////
			// Result
			//                                                            Mbps   CD/DVD  CPU %
			// ---------------- ----------- ---- : ------------- (-------------) ------ ------ */
			// Random R/W 50:50 Read  Speed 55.5           kB/s
			// Random R/W 50:50 Write Speed                kB/s
			// Random R/W 50:50 Read  Count                io/s
			// Random R/W 50:50 Write Count                io/s
			{
				enum res_enum
				{
					RES_RKBS,
					RES_WKBS,
					RES_RIOS,
					RES_WIOS,
					RES_MAX ,
				};
				unsigned	res;

				clk = clkend - clkstart;
				for ( res = 0; res < RES_MAX; res++ )
				{
					char		testmsg[80];
					INT64		speed;
					unsigned	percent;

					switch ( res )
					{
						case RES_RKBS:
							percent = (int)( (10000 * rsize  / (rsize  + wsize )+5)/10 );
							sprintf(testmsg, "Read  Speed %2u.%1u", percent / 10, percent % 10);
							speed = (INT64)CLOCKS_PER_SEC * (INT64)rsize  / (INT64)clk;
							break;
						case RES_WKBS:
							percent = (int)( (10000 * wsize  / (rsize  + wsize )+5)/10 );
							sprintf(testmsg, "Write Speed %2u.%1u", percent / 10, percent % 10);
							speed = (INT64)CLOCKS_PER_SEC * (INT64)wsize  / (INT64)clk;
							break;

						case RES_RIOS:
							percent = (int)( (10000 * rcount / (rcount + wcount)+5)/10 );
							sprintf(testmsg, "Read  Count %2u.%1u",  percent / 10, percent % 10);
							speed = (INT64)CLOCKS_PER_SEC * (INT64)rcount / (INT64)clk;
							break;
						case RES_WIOS:
							percent = (int)( (10000 * wcount / (rcount + wcount)+5)/10 );
							sprintf(testmsg, "Write Count %2u.%1u", percent / 10, percent % 10);
							speed = (INT64)CLOCKS_PER_SEC * (INT64)wcount / (INT64)clk;
							break;
					}

					if ( res == RES_RKBS || res == RES_WKBS )
					{
						char		kbs_n[80], kbs[80], mbps[80], discspeed[16],cpuusage[16];
						int			kiro;
						INT64		ub;

						switch ( this->opt.byteUnit )
						{
							case BYTEUNIT_210:
								kiro = 'K';
								ub = I64(1024);
								break;
							case BYTEUNIT_103:
							default:
								kiro = 'k';
								ub = I64(1000);
								break;
						}
						sprintf(kbs_n, "%u.%1u"
						  , (unsigned)(speed / ub)
						  , (unsigned)((I64(10) * speed / ub) % I64(10))
						);
						sprintf(kbs, "%s %cB/s", kbs_n, kiro);
						sprintf(mbps
						  , "%u.%02u Mbps"
						  , (unsigned)(speed * I64(8)/ (ub*ub))
						  , (unsigned)((I64(100) * speed * I64(8) / (ub*ub)) % I64(100))
						);
						if ( this->ben.discSpeed > 0 )
						{
							char	t1[80];

							sprintf(t1
							  , "x%u.%1u"
							  , (unsigned)(I64(1000) * speed / this->ben.discSpeed / I64(1000))
							  , (unsigned)((I64(10) * I64(1000) * speed / this->ben.discSpeed / I64(1000)) % I64(10))
							);
							sprintf(discspeed, " %6s", t1);
						} else
							discspeed[0] = '\0';
						if ( this->opt.cpuUsage != CPUUSAGE_OFF && res == RES_RKBS )
							sprintf(cpuusage, " %5.1f%%", cpuUsage);
						else
							cpuusage[0] = '\0';
						fprintf(stdout, "%-16s %-16s : %13s (%13s)%s%s\n"
						  , testMsgBuf
						  , testmsg
						  , kbs
						  , mbps
						  , discspeed
						  , cpuusage
						);
						if ( this->opt.benchCsvResult != NULL )
						{
							fprintf(this->ben.csvResult, "%s %s, %s\n"
							  , testMsgBuf
							  , testmsg
							  , kbs_n
							);
						}
					} else
					{	// io/s
						char	ios[80];

						sprintf(ios, "%u io/s", (unsigned)speed );
						fprintf(stdout, "%-16s %-16s : %13s (%13s)\n"
						  , testMsgBuf
						  , testmsg
						  , ios
						  , "-------------"
						);
						if ( this->opt.benchCsvResult != NULL )
						{
							fprintf(this->ben.csvResult, "%s %s, %u\n"
							  , testMsgBuf
							  , testmsg
							  , (unsigned)speed
							);
						}
					}
				} // end of for (res; 

			}
		} // end of for( test;

		retcode = 0;
	}
	catch (int code) { retcode = code; }
	if ( rbuf != NULL )
		delete [] rbuf;
	if ( wbuf != NULL )
		delete [] wbuf;
	if ( rtbl != NULL )
		delete [] rtbl;
	if ( wtbl != NULL )
		delete [] wtbl;
	return retcode;
}

int		DevTest::BenchCommand(DevTest::DevParam & devParam, unsigned benmode)
{
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned		pathId, targetId, lun;
	BYTE 						*buffer = NULL;
	INT64			maxblk;
	unsigned		sectorSize;
	enum test_enum
	{
		TEST_0,
		TEST_1,
		TEST_2,
		TEST_NUM
	};
	static char 	*TestText[] = 
	{
		"Test Unit Ready command",
		"No Motion Seek command" ,
		"Average latency Time"   ,
	};
	int				test;
	int				retcode;
	BOOL			cmdSeek = TRUE, cmdVerify = TRUE;

	if ( this->opt.benchCommand == FALSE )
		return 0;
	if ( devParam.inqData == NULL )
		return -1;
	fprintf(stdout,"\n");
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{
		fprintf(stderr,"ScsiReadCapacity error!\n");
		return -1;
	}
	devParam.capa = &capa;
	maxblk     = this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
	sectorSize = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);

	if ( this->devCtl.ScsiSeek(pathId, targetId, lun, I64(0)) < 0 )
		cmdSeek = FALSE;
	if ( this->devCtl.ScsiVerify(pathId, targetId, lun, I64(0), 1) < 0 )
		cmdVerify = FALSE;
	if ( (buffer = new BYTE [sectorSize]) == NULL )
		return -1;
	try
	{
		fprintf(stdout, "%-33s : %9s %-19s%s\n"
		  , "TEST MODE"
		  , "TIME(ms)"
		  , "REMARK"
		  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " CPU  %"
		);
		fprintf(stdout, "--------------------------------- : --------- -------------------%s\n"
		  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " ------"
		);
		for ( test = 0; test < TEST_NUM; test++)
		{
			Clock_t		clk, clklast, clkstart, clkend;
			unsigned	count;
			BOOL		error;
			char		textAdd[80];

			fprintf(stdout, "%-33s : "
			  , TestText[test]
			);

			{	// wait
				this->devCtl.ScsiSeek(pathId, targetId, lun, I64(0));
				MSLEEP(100);
			}

			textAdd[0] = '\0';
			switch ( test )
			{
				case TEST_0:
					break;
				case TEST_1:
					if ( cmdSeek == FALSE )
						sprintf(textAdd, "(use Read command)");
					break;
				case TEST_2:
					if ( cmdVerify == FALSE )
						sprintf(textAdd, "(use Read Command)");
					break;
			}

			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				this->CollectCpuUsage();
			error = FALSE;
			clkstart = clklast = this->Clock();
			for ( count = 0; count < COMMAND_TEST_COUNT; count++ )
			{
				clk = this->Clock();
				if ( clk < clklast )
				{
					fprintf(stderr,"Clock error!\n");
					throw -1;	// clock error!
				}
				clklast = clk;
				switch ( test )
				{
					case TEST_0:
						if ( this->devCtl.ScsiTestUnitReady(pathId, targetId, lun) < 0 )
						{
							fprintf(stderr,"ScsiTestUnitReady error!\n");
							error = TRUE;
							break;
						}
						break;

					case TEST_1:
						if ( cmdSeek )
						{
							if ( this->devCtl.ScsiSeek(pathId, targetId, lun, I64(0)) < 0 )
							{
								fprintf(stderr,"ScsiSeek error!\n");
								error = TRUE;
								break;
							}
						} else
						{
							if ( this->BenchRead(pathId, targetId, lun, I64(0), 1, sectorSize, buffer) < (int)sectorSize )
							{
								fprintf(stderr,"ScsiRead error!\n");
								error = TRUE;
								break;
							}
						}
						break;

					case TEST_2:
						if ( cmdVerify )
						{
							if ( this->devCtl.ScsiVerify(pathId, targetId, lun, I64(0), 1) < 0 )
							{
								fprintf(stderr,"ScsiVerify error!\n");
								error = TRUE;
								break;
							}
						} else
						{
							if ( this->BenchRead(pathId, targetId, lun, I64(0), 0, 0, NULL) < 0 )
							{
								fprintf(stderr,"ScsiRead error!\n");
								error = TRUE;
								break;
							}
						}
						break;
				}
				if ( error )
					break;
			}
			if ( error == FALSE )
			{
				char	ms[80];

				clkend = this->Clock();
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					this->CollectCpuUsage();
				clk = clkend - clkstart;
				sprintf(ms, "%u.%02u"
				  , (unsigned)(   I64(1000) * (INT64)clk / (INT64)COMMAND_TEST_COUNT / (INT64)CLOCKS_PER_SEC)
				  , (unsigned)((I64(100000) * (INT64)clk / (INT64)COMMAND_TEST_COUNT / (INT64)CLOCKS_PER_SEC) % I64(100))
				);
				fprintf(stdout, "%7sms %-19s", ms, textAdd );
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					fprintf(stdout, " %5.1f%%", this->GetCpuUsage() );
				if ( this->opt.benchCsvResult != NULL )
				{
					fprintf(this->ben.csvResult, "%-33s , %s\n"
					  , TestText[test]
					  , ms
					);
				}
			}
			fprintf(stdout, "\n");
		} // end of for(test;;)
		retcode = 0;
	}
	catch ( int ret ) { retcode = ret; }
	if ( buffer != NULL )
		delete [] buffer;

	return retcode;
}

int		DevTest::BenchSeek(DevTest::DevParam & devParam, unsigned benmode)
{
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
	BYTE 						*buffer = NULL;
	enum test_enum
	{
		TEST_SEQUENTION,
		TEST_RANDOM    ,
		TEST_NUM
	};
	static char 	*TestText[] = 
	{
		"Sequential Seek",
		"Random Seek"    ,
	};
	INT64			maxblk;
	unsigned		sectorSize, seekBlks = 0;
	int				test;
	int				retcode;
	BOOL			cmdSeek = TRUE, cmdVerify = TRUE;

	if ( this->opt.benchSeek == FALSE )
		return 0;
	if ( devParam.inqData == NULL )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capa), &capa) < 0 )
	{
		fprintf(stderr,"ScsiReadCapacity error!\n");
		return -1;
	}
	fprintf(stdout, "\n");
	devParam.capa = &capa;
	maxblk     = this->devCtl.ScsiGetLBA(capa.logicalBlockAddress);
	sectorSize = (unsigned)this->devCtl.ScsiGetLBA(capa.blockLength);
	seekBlks   = this->ben.seekSector;
	if ( seekBlks == 0 )
	{
#if	defined(WIN32)
		DISK_GEOMETRY	geo;

		if ( this->devCtl.DiskGetDriveGeometry(geo) == 0 )
			seekBlks = (unsigned)geo.SectorsPerTrack;
		else
			seekBlks = 32;
#endif
#if	defined(LINUX)
		seekBlks = 32;
#endif
	}
	if ( seekBlks > maxblk )
		seekBlks = (unsigned)(maxblk / I64(100));

	if ( this->devCtl.ScsiSeek(pathId, targetId, lun, I64(0)) < 0 )
		cmdSeek = FALSE;
	if ( this->devCtl.ScsiVerify(pathId, targetId, lun, I64(0), 1) < 0 )
		cmdVerify = FALSE;
	if ( (buffer = new BYTE [sectorSize]) == NULL )
		return -1;

	try
	{
		fprintf(stdout, "%-33s : %9s %-19s%s\n"
		  , "TEST MODE"
		  , "TIME(ms)"
		  , "REMARK"
		  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " CPU  %"
		);
		fprintf(stdout, "--------------------------------- : --------- -------------------%s\n"
		  , this->opt.cpuUsage == CPUUSAGE_OFF ? "" : " ------"
		);
		for ( test = 0; test < TEST_NUM; test++)
		{
			Clock_t		clk, clklast, clkstart, clkend;
			INT64		addr;
			unsigned	count;
			BOOL		error;
			char		textAdd[80];

			{
				char	text[80];

				switch ( test )
				{
					case TEST_SEQUENTION:
						sprintf(text, "%s (seek sector %u)", TestText[test], seekBlks);
							break;

					case TEST_RANDOM    :
						sprintf(text, "%s", TestText[test]);
						break;
				}
				fprintf(stdout, "%-33s : ", text);
			}

			{	// wait
				this->devCtl.ScsiSeek(pathId, targetId, lun, I64(0));
				MSLEEP(100);
			}

			textAdd[0] = '\0';
			if ( cmdSeek == FALSE )
				sprintf(textAdd, "(use Read Command)");
			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				this->CollectCpuUsage();
			error = FALSE;
			addr = 0;
			count = 0;
			clkstart = clklast = this->Clock();
			clkend = clkstart + this->ben.testTimeSec * CLOCKS_PER_SEC;
			while ( (clk = this->Clock()) < clkend )
			{
				if ( clk < clklast )
				{
					fprintf(stderr,"Clock error!\n");
					throw -1;	// clock error!
				}
				clklast = clk;

				if ( cmdSeek )
				{
					if ( this->devCtl.ScsiSeek(pathId, targetId, lun, addr) < 0 )
					{
						fprintf(stderr,"ScsiSeek error!\n");
						error = TRUE;
						break;
					}
				} else
				{
					if ( this->BenchRead(pathId, targetId, lun, addr, 1, sectorSize, buffer) < (int)sectorSize )
					{
						fprintf(stderr,"ScsiRead error!\n");
						error = TRUE;
						break;
					}
				}
				count++;
				switch ( test )
				{
					case TEST_SEQUENTION:
						addr += (INT64)seekBlks;
						if ( addr >= maxblk )
							addr = 0;
						break;
					case TEST_RANDOM:
						do
						{
							addr = maxblk * (INT64)(rand() % RANDOM_SCALE) / (INT64)RANDOM_SCALE;
						} while ( addr >= maxblk );
						break;
				}
			}
			if ( error )
				break;
			clkend = this->Clock();
			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				this->CollectCpuUsage();
			if ( error == FALSE && count > 0 )
			{
				char	ms[80];

				clkend = this->Clock();
				clk = clkend - clkstart;
				sprintf(ms, "%u.%02u"
				  , (unsigned)(   I64(1000) * (INT64)clk / (INT64)count / (INT64)CLOCKS_PER_SEC)
				  , (unsigned)((I64(100000) * (INT64)clk / (INT64)count / (INT64)CLOCKS_PER_SEC) % I64(100))
				);
				fprintf(stdout, "%7sms %-19s", ms, textAdd );
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					fprintf(stdout, " %5.1f%%", this->GetCpuUsage() );
				if ( this->opt.benchCsvResult != NULL )
				{
					fprintf(this->ben.csvResult, "%-33s , %s\n"
					  , TestText[test]
					  , ms
					);
				}
			}
			fprintf(stdout, "\n");
		} // end of for(test;;)
		retcode = 0;
	}
	catch ( int ret ) { retcode = ret; }
	if ( buffer != NULL )
		delete [] buffer;

	return retcode;
}

int		DevTest::BenchRead(int pathId, int targetId, int lun, INT64 addr, unsigned count, unsigned size, void * buffer)
{
#if	defined(WIN32)
	if ( this->ben.useWinFile )
	{	// need devCtl.CreateCtlWork
		LARGE_INTEGER	pos;
		DWORD			dwResult;

		pos.QuadPart = addr;
		if ( pos.LowPart = ::SetFilePointer(this->devCtl.GetHandle(), pos.LowPart, &pos.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER
		  && ::GetLastError() != NO_ERROR )
			return -1;
		if ( ::ReadFile(this->devCtl.GetHandle(), buffer, size, &dwResult, NULL) == FALSE )
			return -1;
		return (int)dwResult;
	} else
#endif
	{
		return this->devCtl.ScsiRead(pathId, targetId, lun, addr, count, size, buffer);
	}
}

