/////////////////////////////////////////////////////////////////////////
//	"devtest_blkcopy.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_blkcopy.cpp 126 2004-07-12 10:39:31Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<time.h>
#include	"devtest.h"
#include	"defs.h"

struct Devwork
{
	DevCtl						devCtl;
	INT64						maxBlk;
	unsigned					sectorSize;
	DevCtl::SCSICAPACITYDATA	capa;
	unsigned					pathId, targetId, lun;
};

static	int	SetBlockCopyWork(Devwork & work, DevTest::DevParam & devParam)
{
	work.pathId   = devParam.inqData->GetPathId();
	work.targetId = devParam.inqData->GetTargetId();
	work.lun      = devParam.inqData->GetLun();
	if ( work.devCtl.ScsiReadCapacity( work.pathId, work.targetId, work.lun, sizeof(work.capa), &work.capa) < 0 )
		return -1;
	devParam.capa = &work.capa;
	work.maxBlk     =           work.devCtl.ScsiGetLBA(work.capa.logicalBlockAddress);
	work.sectorSize = (unsigned)work.devCtl.ScsiGetLBA(work.capa.blockLength);
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Block Copy

int		DevTest::BlockCopy(DevTest::DevParam & srcDev, DevTest::DevParam & dstDev)
{
	DevTargetParam	srctg, dsttg;
	int				res1, res2;
	BYTE			*buffer = NULL;
	Devwork			src, dst;
	int				retcode = 0;
	time_t						tmStart, tmEnd, tmNow, tmLast;
	Clock_t						clkLast, clk;

	fprintf(stdout,"***** Block Copy *****\n\n");

	try
	{
		INT64		rwblks = I64(0);
		INT64		srcLba, dstLba, copyBlk;
		unsigned	count;

		res1 = this->OpenTargetDevice(src.devCtl, srctg, srcDev);
		res2 = this->OpenTargetDevice(dst.devCtl, dsttg, dstDev);
		if ( res1 < 0 || res2 < 0 )
			throw -1;

		if ( ::SetBlockCopyWork(src, srcDev) < 0 )
			throw -1;
		if ( ::SetBlockCopyWork(dst, dstDev) < 0 )
			throw -1;

		fprintf(stdout, "  Read  Device %s %13"FI64"u\n", srcDev.desc, src.maxBlk);
		fprintf(stdout, "  Write Device %s %13"FI64"u\n", dstDev.desc, dst.maxBlk);

		if ( src.sectorSize != dst.sectorSize )
		{
			fprintf(stderr, "Sector size mismatch! %u != %u\n", src.sectorSize, dst.sectorSize);
			throw -1;
		}
		if ( (buffer = new BYTE [this->ben.maxBufferSize]) == NULL )
		{
			fprintf(stderr, "Out of memory!\n");
			throw -1;
		}

		copyBlk = src.maxBlk - this->opt.readAddr;
		if (  copyBlk > dst.maxBlk - this->opt.writeAddr )
			copyBlk = dst.maxBlk - this->opt.writeAddr;
		if ( this->opt.count > I64(0) && copyBlk > this->opt.count )
			copyBlk = this->opt.count;

		if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		{
			this->InitCpuUsage();
			this->CollectCpuUsage();
		}

		time(&tmStart);
		tmLast = tmStart;
		clkLast = this->Clock();

		srcLba = this->opt.readAddr;
		dstLba = this->opt.writeAddr;
		count = 0;
		while ( rwblks < copyBlk )
		{
			unsigned	rcount, wcount;
			unsigned	rsize, wsize;

			time(&tmNow);
			if ( difftime(tmNow, tmLast) > 1 )
			{
				char		speed[80], remain[80], cpuusage[80];
				unsigned	total_sec_per_blk, sec_per_blk, remSec;

				speed[0] = remain[0] = 0;
				clk = this->Clock();
				if ( clk - clkLast > 0 )
				{
					sec_per_blk = (unsigned)( (INT64)CLOCKS_PER_SEC * (INT64)count / (INT64)(clk - clkLast) );
					sprintf(speed, "%u.%1ukB/s"
					  , (unsigned) ( (INT64)sec_per_blk * (INT64)src.sectorSize / I64(1000))
					  , (unsigned) (I64(10) * (INT64)sec_per_blk * (INT64)src.sectorSize / I64(1000)) % 10
					);
					total_sec_per_blk = (unsigned)( rwblks / (INT64)(difftime(tmNow, tmStart)) );
					if ( sec_per_blk > 0 && total_sec_per_blk > 0 )
					{
						sec_per_blk = (sec_per_blk + total_sec_per_blk) / 2;
						remSec = (unsigned)((copyBlk - rwblks) / (INT64)sec_per_blk);
						this->FormatTimeRemain(remain, remSec);
					}
				} else
				{
					strcpy(speed, "?.?");
					strcpy(remain, "????");
				}
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				{
					this->CollectCpuUsage();
					sprintf(cpuusage, ":CPU%5.1f%%", (double)this->GetCpuUsage() );
				} else
					cpuusage[0] = '\0';
				fprintf(stderr, "%-6s%10"FI64"u/%"FI64"u(%3u.%1u%%)%14s%-9s(Remain%13s\r"
				  , "Copy"
				  , (INT64)rwblks
				  , (INT64)copyBlk
				  , (unsigned)(I64(100) * rwblks / copyBlk)
				  , (unsigned)(I64(1000) * rwblks / copyBlk) % 10
				  , speed
				  , cpuusage
				  , remain
				);
				tmLast = tmNow;
				clkLast = clk;
				count = 0;
			}

			// Read
			if ( srcLba >= src.maxBlk )
			{
				fprintf(stderr, "Read LBA over!\n");
				break;
			}
			rcount = this->ben.maxBufferSize / src.sectorSize;
			if ( srcLba + (INT64)rcount >= src.maxBlk )
				rcount = (unsigned)(src.maxBlk - srcLba);
			rsize = rcount * src.sectorSize;
			if ( src.devCtl.ScsiRead(src.pathId, src.targetId, src.lun, srcLba, rcount, rsize, buffer) < (int)rsize )
			{
				fprintf(stderr, "ScsiRead Error! (lba %"FI64"u)\n", srcLba);
				break;
			}
			srcLba += (INT64)rcount;

			// Write
			if ( dstLba >= dst.maxBlk )
			{
				fprintf(stderr, "Write LBA over!\n");
				break;
			}
			wcount = this->ben.maxBufferSize / dst.sectorSize;
			if ( dstLba + (INT64)wcount >= dst.maxBlk )
				wcount = (unsigned)(dst.maxBlk - dstLba);
			wsize = wcount * dst.sectorSize;
			// Delete Source Code (ScsiWrite)
			dstLba += (INT64)wcount;

			rwblks += (INT64)wcount;
			count += wcount;
		}

		time(&tmEnd);
		{
			unsigned	sec;
			char		tmbuf[32];

			sec = (unsigned) difftime(tmEnd, tmStart);
			fprintf(stdout, "\nBlock Copy completed. %usec, (%13s)\n"
			  , sec, this->FormatTimeRemain(tmbuf, sec)
			);
		}

		retcode = 0;
	} catch (int code) { retcode = code; }
	this->CloseTargetDevice(src.devCtl, srctg, srcDev);
	this->CloseTargetDevice(dst.devCtl, dsttg, dstDev);
	if ( buffer != NULL )
		delete [] buffer;
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		this->TermCpuUsage();

	return 0;
}


