/////////////////////////////////////////////////////////////////////////
//	"udflib_lib.cpp"
//------------------------------------------------------------------------
//	$Id: udflib_lib.cpp 50 2003-12-11 01:39:41Z  $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<string.h>
#include	<stdio.h>
#include	"udflib.h"

void	UdfLib::VRSInfo::Clear()
{
	Vrs	*vrs, *vrs_next;

	if ( (vrs = this->vrsTop) != NULL )
	{
		while ( vrs != NULL )
		{
			vrs_next = vrs->next;
			delete vrs;
			vrs = vrs_next;
		}
	}
	this->vrsTop = this->vrsLast = NULL;
	this->numVrs = 0;
}

int		UdfLib::VRSInfo::Add(unsigned type, unsigned loc, UdfLib::VRS * pvrs)
{
	Vrs		*vrs;

	if ( (vrs = new Vrs) == NULL )
		return -1;
	vrs->next = NULL;
	vrs->type = type;
	vrs->loc  = loc;
	memcpy(&vrs->vrs, pvrs, sizeof(vrs->vrs));
	if ( this->vrsLast == NULL )
	{
		this->vrsTop = this->vrsLast = vrs;
	} else
	{
		this->vrsLast->next = vrs;
		this->vrsLast = vrs;
	}
	this->numVrs++;

	return 0;
}

void	UdfLib::VolDesInfo::Clear()
{
	VolDes	*voldes, *voldes_next;

	if ( (voldes = this->desTop) != NULL )
	{
		while ( voldes != NULL )
		{
			voldes_next = voldes->next;
			delete voldes;
			voldes = voldes_next;
		}
	}
	this->desTop = this->desLast = NULL;
	this->numDes = 0;
}

int		UdfLib::VolDesInfo::AddDes(unsigned type, unsigned loc, unsigned len, unsigned size, UdfLib::DTag * tag)
{
	VolDes	*voldes;

	if ( (voldes = (VolDes *)new BYTE [sizeof(VolDes) + size]) == NULL )
		return -1;
	voldes->next = NULL;
	voldes->type = type;
	voldes->loc  = loc;
	voldes->len  = len;
	voldes->des  = (UdfLib::VolDes::Des *)(voldes + 1);
	memcpy(voldes->des, tag, size);
	if ( this->desLast == NULL )
	{
		this->desTop = this->desLast = voldes;
	} else
	{
		this->desLast->next = voldes;
		this->desLast = voldes;
	}
	this->numDes++;

	return 0;
}

int		UdfLib::ReadSector(INT64 adr, unsigned count, unsigned size, void * buffer)
{
	if ( this->devCtl == NULL || this->inqData == NULL )
		return -1;
	if ( this->devCtl->ScsiRead(
	   this->inqData->GetPathId(),
	  this->inqData->GetTargetId(),
	  this->inqData->GetLun(),
	  this->sec_off + adr,
	  count,
	  size,
	  buffer ) < (int)size )
		return -1;
	return 0;
}

int		UdfLib::GetBlkSize(void)
{
	if ( this->devCtl == NULL || this->capa == NULL )
		return -1;
	return (int)this->devCtl->ScsiGetLBA(capa->blockLength);
}

INT64	UdfLib::GetBlkNum(void)
{
	if ( this->devCtl == NULL || this->capa == NULL )
		return -1;
	return this->sec_num;
//	return this->devCtl->ScsiGetLBA(capa->logicalBlockAddress);
}

int 	UdfLib::CheckTag(UdfLib::DTag & tag, Uint32 tagLoc, Uint16 tagId, int crc_min, int crc_max)
{
	Uint8		checksum;
	int			i;
	Uint16		crc;

	checksum = 0;
	for ( i = 0; i < 4; i++)
		checksum += *((Uint8 *)&tag + i);
	for ( i = 5; i < 16; i++)
		checksum += *((Uint8 *)&tag + i);

	if ( tag.tagChecksum != checksum )
	{
		fprintf(stdout,"Checksum error!\n");
		return -1;	// checksum error
	}

	if ( (tagId != (Uint16)-1) && (tagId != this->U_endian16(tag.tagID)) )
	{
		fprintf(stdout,"tag id error!\n");
		return -1;
	}

	if (( this->U_endian16(tag.crcLen) >= crc_min) && (this->U_endian16(tag.crcLen) <= crc_max))
	{
		crc = this->doCrc((Uint8 *)&tag + 16, this->U_endian16(tag.crcLen));
		if ( crc != this->U_endian16(tag.descriptorCrc))
		{
			fprintf(stdout,"crc error!\n");
			return -1;
		}
	} else
	{
		fprintf(stdout,"crc len error!\n");
		return -1;
	}
 
	if ( tagLoc != this->U_endian32(tag.tagLoc) )
	{
		fprintf(stdout,"tag loc error!\n");
		return -1;
	}

#if 0
	if (Version_OK && (this->U_endian16(tag.descriptorVersion) != UDF_Version))
		return -1;
	if ( Serial_OK && (this->U_endian16(tag.tagSerialNum != Serial_No)))
		return -1;
#endif

	return 0;
}

#define MASK 0x1021

UdfLib::Uint16	UdfLib::doCrc(UdfLib::Uint8 * buffer, int n)
{
	Uint16	crc = 0;
	int		byte, bit;
	Uint8	bitval, msb;

	if ( n > 2032)
	{
		crc = 0xFFFF;
	} else
	{
		for ( byte = 0; byte < n; byte++ )
		{
			for ( bit = 7; bit >= 0; bit--)
			{
				bitval = (*(buffer + byte) >> bit) & 0x01;
				msb = crc >> 15;
				crc <<=  1;
				crc ^= MASK * (bitval ^ msb);
			}
		}
	}
	return crc;
}


int UdfLib::track_volspace(UdfLib::VolDesInfo volDesInfo, Uint32 location, Uint32 length, char *name)
{
	int i, j, error;

	error = 0;
	if ( length == 0 )
		return 0;

    //point to the location where the new extent is >= the list member
	for (i = 0; (i < (int)volDesInfo.volspcListLen) && (location > volDesInfo.volspc[i].location); i++)
		;
	if ( i > 0 )
	{
		if (((location + length - 1) >= volDesInfo.volspc[i-1].location) && 
		  (location < (volDesInfo.volspc[i-1].location + volDesInfo.volspc[i-1].length)))
		{
		//	Error.Code = ERR_VOL_SPACE_OVERLAP;
		//	Error.Sector = Location;
			error = TRUE;
		}
	}
	if ( i < (int)volDesInfo.volspcListLen )
	{
		if (((location + length - 1) >= volDesInfo.volspc[i].location) && 
		  (location < (volDesInfo.volspc[i].location + volDesInfo.volspc[i].length)))
		{
		//	Error.Code = ERR_VOL_SPACE_OVERLAP;
		//	Error.Sector = Location;
			error = TRUE;
		}
	}
	if ( volDesInfo.volspcListLen < MAX_VOL_EXTS - 1)
	{
		for (j = volDesInfo.volspcListLen; j > i; j--)
		{
			volDesInfo.volspc[j] = volDesInfo.volspc[j - 1];
		}
		volDesInfo.volspc[i].location = location;
		volDesInfo.volspc[i].length   = length;
		volDesInfo.volspc[i].name     = name;
		volDesInfo.volspcListLen++;
	} else
	{
	//	printf("**Too many volume extents.");
	}
	return error ? -1 : 0;
}


int		UdfLib::CheckRegid(UdfLib::udfEntityId * reg, char * id)
{
	int		error = 0;
	Uint8	cbuf[23];

	strncpy((char *)cbuf, id, 23);
	if ( reg == NULL )
		return -1;
	if ( strncmp((char *)reg->id, (char *)cbuf, 23) != 0 )
		return -1;
	if (reg->OSClass > 6)
		return -1;
	if (( this->U_endian16(reg->UDFRevision) < 0x100) || (this->U_endian16(reg->UDFRevision) > 0x200))
		return -1;
	return 0;
}

int UdfLib::UncompressUnicode(int numberOfBytes, UdfLib::Uint8 *UDFCompressed, UdfLib::unicode_t *unicode)
{
	unsigned int compID;
	int returnValue, unicodeIndex, byteIndex;

	/* Use UDFCompressed to store current byte being read. */
	compID = UDFCompressed[0];
	/* First check for valid compID. */
	if ( compID != 8 && compID != 16 )
	{
		returnValue = -1;
	} else
	{
		unicodeIndex = 0;
		byteIndex = 1;
		/* Loop through all the bytes. */
		while ( byteIndex < numberOfBytes )
		{
			if ( compID == 16 )
			{	/* Move the first byte to the high bits of the unicode char. */
				unicode[unicodeIndex] = UDFCompressed[byteIndex++] << 8;
			} else
			{
				unicode[unicodeIndex] = 0;
			}
			if ( byteIndex < numberOfBytes )
			{	/* Then the next byte to the low bits. */
				unicode[unicodeIndex] |= UDFCompressed[byteIndex++];
			}
			unicodeIndex++;
		}
		returnValue = unicodeIndex;
	}
	return returnValue;
}

