//////////////////////////////////////////////////////////////////////////
//	"devctl.cpp"
//------------------------------------------------------------------------
//	$Id: devctl.cpp 30 2003-12-09 07:39:09Z  $
//////////////////////////////////////////////////////////////////////////

#include	"devctl.h"
#include	<tchar.h>
#include	<stdio.h>

DevCtl::DevCtl()
{
	this->devname = NULL;
	this->handle = NULL;
	this->htype = HT_UNKNOWN;
#if defined(DEVCTL_USE_FILEIO)
	this->ctl = NULL;
#endif
	this->scsi_timeout = SCSI_TIMEOUT;
}

DevCtl::~DevCtl()
{
	if ( this->handle != NULL )
	{
		::CloseHandle(this->handle);
		this->handle = NULL;
	}
	if ( this->devname != NULL )
	{
		delete [] this->devname;
		this->devname = NULL;
	}
#if defined(DEVCTL_USE_FILEIO)
	if ( this->ctl != NULL )
	{
		delete this->ctl;
		this->ctl = NULL;
	}
#endif
}

int		DevCtl::OpenDevice(const TCHAR *name, unsigned rwmode)
{
	DWORD	dwAcc, dwShare, dwAtt;

	if ( this->handle != NULL )
		return -1;
	if ( this->devname != NULL )
		delete [] this->devname;
	if ( (this->devname= new TCHAR [_tcslen(name)+1]) == NULL )
		return -1;
	_tcscpy(this->devname, name);

	this->htype = HT_UNKNOWN;
	if ( _tcsnicmp(this->devname, TEXT("\\\\.\\PhysicalDrive"), 17) == 0 )
		this->htype = HT_DISK;
	else if ( _tcsnicmp(this->devname, TEXT("\\\\.\\"), 3) == 0
	  && isalpha(this->devname[4]) && this->devname[5] == ':' && this->devname[6] == '\0' )
		this->htype = HT_DRIVE;
	else if ( _tcsnicmp(this->devname, TEXT("\\\\.\\scsi"), 7) == 0 )
		this->htype = HT_SCSI;

	dwAcc = dwShare = dwAtt = 0;
	if ( rwmode & RW_READ )
		dwAcc |= GENERIC_READ;
	if ( rwmode & RW_WRITE )
		dwAcc |= GENERIC_WRITE;
	if ( rwmode & RW_SHARE_READ )
		dwShare |= FILE_SHARE_READ;
	if ( rwmode & RW_SHARE_WRITE )
		dwShare |= FILE_SHARE_WRITE;

	if ( (this->handle = ::CreateFile(devname, dwAcc, dwShare, NULL, OPEN_EXISTING, dwAtt, NULL)) == INVALID_HANDLE_VALUE )
	{
		this->handle = NULL;
		return -1;
	}

	return 0;
}

int		DevCtl::CloseDevice(void)
{
#if defined(DEVCTL_USE_FILEIO)
	if ( this->ctl != NULL )
	{
		delete this->ctl;
		this->ctl = NULL;
	}
#endif
	if ( this->handle == NULL )
		return -1;
	::CloseHandle(this->handle);
	this->handle = NULL;
	return 0;
}

int		DevCtl::GetFileType(void)
{
	if ( this->handle == NULL )
		return -1;
	return ::GetFileType(this->handle);
}

int		DevCtl::GetFileSize(LARGE_INTEGER & size)
{
	if ( this->handle == NULL )
		return -1;
	if ( size.LowPart = ::GetFileSize(this->handle, (DWORD *)&size.HighPart) == -1
	  && ::GetLastError() != NO_ERROR )
		return -1;
	return 0;
}

DevCtl::INT64	DevCtl::GetFileSize(void)
{
	LARGE_INTEGER	size;

	if ( this->GetFileSize(size) < 0 )
		return -1I64;
	return size.QuadPart;
}

int		DevCtl::SetFilePointer(LARGE_INTEGER & pos, int method)
{
	if ( this->handle == NULL )
		return -1;
	if ( pos.LowPart = ::SetFilePointer(this->handle, pos.LowPart, &pos.HighPart, (DWORD)method) == INVALID_SET_FILE_POINTER
	  && ::GetLastError() != NO_ERROR )
		return -1;
	return 0;
}
int		DevCtl::SetFilePointer(INT64 & pos, int method)
{
	LARGE_INTEGER	filepos;

	filepos.QuadPart = pos;
	if ( this->SetFilePointer(filepos, method) < 0 )
		return -1;
	pos = filepos.QuadPart;
	return 0;
}

int		DevCtl::DeviceIoControl(DWORD ioctl, void * inbuffer, unsigned insize, void * outbuffer, unsigned outsize, DWORD * byteRet)
{
	DWORD	dw;

	if ( this->handle == NULL )
		return -1;
	if ( byteRet == NULL )
		byteRet = &dw;
	return (int) ::DeviceIoControl( this->handle, ioctl,
	  inbuffer , insize,	// input
	  outbuffer, outsize,	// output
	  (DWORD *) byteRet, NULL);
}

int		DevCtl::DiskIsWritable(void)
{
	return this->DeviceIoControl(IOCTL_DISK_IS_WRITABLE,  NULL, 0,  NULL, 0);
}

int		DevCtl::DiskGetDriveGeometry(DISK_GEOMETRY & geo)
{
	DWORD		dwByteRet;

	if ( this->DeviceIoControl( IOCTL_DISK_GET_DRIVE_GEOMETRY,
	  NULL, 0,						// input
	  (void *) &geo, sizeof(geo),	// output
	  &dwByteRet) == FALSE )
		return -1;	// error

	return 0;
}

int		DevCtl::StorageGetMediaTypes(DISK_GEOMETRY & geo)
{
	DWORD		dwByteRet;

	if ( this->DeviceIoControl( IOCTL_STORAGE_GET_MEDIA_TYPES,
	  NULL, 0,						// input
	  (void *) &geo, sizeof(geo),	// output
	  &dwByteRet) == FALSE )
		return -1;	// error

	return 0;
}

int		DevCtl::CdromGetDriveGeometry(DISK_GEOMETRY & geo)
{
	DWORD		dwByteRet;

	if ( this->DeviceIoControl( IOCTL_CDROM_GET_DRIVE_GEOMETRY,
	  NULL, 0,						// input
	  (void *) &geo, sizeof(geo),	// output
	  &dwByteRet) == FALSE )
		return -1;	// error

	return 0;
}


#if (_WIN32_WINNT >= 0x500)
int		DevCtl::DiskGetDriveGeometryEx(DISK_GEOMETRY_EX & geo)
{
	DWORD		dwByteRet;

	if ( this->DeviceIoControl(IOCTL_DISK_GET_DRIVE_GEOMETRY_EX,
	  NULL, 0,						// input
	  (void *) &geo, sizeof(geo),	// output
	  &dwByteRet) == FALSE )
		return -1;	// error

	return 0;
}
#endif

int		DevCtl::StorageGetHotPlugInfo(STORAGE_HOTPLUG_INFO & hi)
{
	DWORD		dwByteRet;

	if ( this->DeviceIoControl(IOCTL_STORAGE_GET_HOTPLUG_INFO,
	  NULL, 0,					// input
	  (void *) &hi, sizeof(hi),	// output
	  &dwByteRet) == FALSE )
		return -1;	// error

	return 0;
}


const TCHAR *	DevCtl::GetMediaTypeName(int mediatype) // static
{
	const TCHAR	   *text;

	switch ( mediatype )
	{
		case F5_1Pt2_512:
		case F5_360_512:
		case F5_320_512:
		case F5_320_1024:
		case F5_180_512:
		case F5_160_512:
		case F5_640_512:
		case F5_720_512:
		case F5_1Pt23_1024:
			text = TEXT("5.25 floppy");
			break;

		case F3_1Pt44_512:
		case F3_2Pt88_512:
		case F3_20Pt8_512:
		case F3_720_512:
		case F3_120M_512:
		case F3_640_512:
		case F3_1Pt2_512:
		case F3_1Pt23_1024:
		case F3_128Mb_512:
		case F3_230Mb_512:
		case F3_200Mb_512:
		case F3_240M_512:
		case F3_32M_512:
			text = TEXT("3.5 floppy");
			break;

		case F8_256_128:
			text = TEXT("8 floppy");
			break;

		case RemovableMedia:
			text = TEXT("Removable");
			break;

		case FixedMedia:
			text = TEXT("Fixed");
			break;

		case Unknown:
		default:
			text = TEXT("Unknown");
			break;
		
	}
	return text;
}

#if defined(DEVCTL_USE_FILEIO)
int		DevCtl::CreateCtlWork(void)
{
	DISK_GEOMETRY	geo;
	int				devtype = -1;

	if ( this->handle == NULL )
		return -1;
	if ( this->ctl != NULL )
		return -1;
	if ( this->DiskGetDriveGeometry(geo) == 0 )
	{
	} else if ( this->StorageGetMediaTypes(geo) == 0 )
	{
	} else if ( this->CdromGetDriveGeometry(geo) == 0 )
	{
		devtype = DEV_CDROM;
	} else
		return -1;

	if ( (this->ctl = new CtlWork) == NULL )
		return -1;
	this->ctl->geo = geo;
	this->ctl->maxBlk = geo.Cylinders.QuadPart * (INT64)geo.TracksPerCylinder * (INT64)geo.SectorsPerTrack;
	this->ctl->sectorSize = geo.BytesPerSector;
	if ( devtype < 0 )
	{
		switch ( geo.MediaType )
		{
			case F5_1Pt2_512:
			case F5_360_512:
			case F5_320_512:
			case F5_320_1024:
			case F5_180_512:
			case F5_160_512:
			case F5_640_512:
			case F5_720_512:
			case F5_1Pt23_1024:
			case F3_1Pt44_512:
			case F3_2Pt88_512:
			case F3_20Pt8_512:
			case F3_720_512:
			case F3_120M_512:
			case F3_640_512:
			case F3_1Pt2_512:
			case F3_1Pt23_1024:
			case F3_128Mb_512:
			case F3_230Mb_512:
			case F3_200Mb_512:
			case F3_240M_512:
			case F3_32M_512:
			case F8_256_128:
				devtype = DEV_FLOPPY;
				break;

			case RemovableMedia:
				devtype = DEV_REMOVABLE;
				break;

			case FixedMedia:
				devtype = DEV_FIXED;
				break;

			case Unknown:
			default:
				devtype = DEV_UNKNOWN;
				break;
		}
	}
	this->ctl->devtype = devtype;

	return 0;
}

int		DevCtl::ReadBlk(INT64 pos, unsigned count, void * buffer)
{
	DWORD	dwResult;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	if ( pos < 0 || pos >= this->ctl->maxBlk )
		return -1;
	if ( this->SetFilePointer(pos, FPOS_BEGIN) < 0 )
		return -1;
	if ( ::ReadFile(this->handle, buffer, this->ctl->sectorSize * count, &dwResult, NULL) == FALSE )
		return -1;
	return (int) dwResult / this->ctl->sectorSize;
}

int		DevCtl::WriteBlk(INT64 pos, unsigned count, const void * buffer)
{
	DWORD	dwResult;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	if ( pos < 0 || pos >= this->ctl->maxBlk )
		return -1;
	if ( this->SetFilePointer(pos, FPOS_BEGIN) < 0 )
		return -1;
	if ( ::WriteFile(this->handle, buffer, this->ctl->sectorSize * count, &dwResult, NULL) == FALSE )
		return -1;
	return (int) dwResult / this->ctl->sectorSize;
}


int		DevCtl::GetDiskFormat(void)
{
	BYTE		   *buffer = NULL;
	int				retcode;

	if ( this->handle == NULL || this->ctl == NULL )
		return -1;
	try
	{
		if ( (buffer = new BYTE [this->ctl->sectorSize]) == NULL )
			throw -1;
		if ( this->ReadBlk( 0I64, 1, buffer) < 1 )
			throw -1;

		if ( buffer[this->ctl->sectorSize - 2] == 0x55 && buffer[this->ctl->sectorSize-1] == 0xAA )	// IBM
		{
			if ( (buffer[0] == 0xE9 || buffer[0] == 0xEB)
			  &&  *(WORD *)(buffer + 0x0B) == this->ctl->sectorSize 
			  && (buffer[0x015] == 0xF0 || buffer[0x015] == 0xF8) )
			{
				throw (int)DF_IBM_FD;
			} else
			{
				throw (int)DF_IBM_HD;
			}
		} else if ( buffer[0] == 'I' && buffer[1] == 'P' && buffer[3] == 'L' )	// FMR HDD
		{
			throw (int)DF_FMR_HD;
		} else
		{
			throw (int)DF_UNKNOWN;
		}
	}
	catch (int code) { retcode = code; }
	if ( buffer != NULL )
		delete [] buffer;
	return retcode;
}
#endif
